# -*- coding: utf-8 -*-
"""
    kaylee.storage
    ~~~~~~~~~~~~~~

    This module provides Kaylee storages' interfaces.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

from abc import ABCMeta, abstractmethod, abstractproperty


class TemporalStorage(object):
    """The interface for applications' temporal results storage.

    Note that using this storage is to be decided by a controller.
    A controller may not need a temporal storage at all.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def add(self, task_id, node_id, result):
        """Stores the task result returned by a node.

        :param task_id: task id
        :param node_id: the id of the node which solved the task
        :param result: task result
        :type task_id: :class:`str`
        :type node_id: :class:`NodeID`
        """

    @abstractmethod
    def remove(self, task_id, node_id=None):
        """Removes a particular task result returned by a defined node from
        the storage."""

    @abstractmethod
    def clear(self):
        """Removes all results from the storage."""

    @abstractmethod
    def __getitem__(self, task_id):
        """Returns the task results.

        :rtype:  ``(node_id : result)`` ... iterator, where
                 ``node_id`` is an instance of :class:`NodeID`.
        """

    @abstractmethod
    def contains(self, task_id, node_id=None, result=None):
        """Checks if any of the task results or a result from a node or a
        particular task result is contained in the storage"""
        pass

    @abstractproperty
    def count(self):
        """The amount of the stored results' unique task id's. This is
        the same as ``len(list(ts.keys()))`` where ``ts`` is an instance of
        :class:`TemporalStorage`."""
        pass

    @abstractproperty
    def total_count(self):
        """The total amount of the stored results. This is the same as
        ``len(sum(nr) for nr in ts.values())`` where ``ts`` is an instance
        of  `TemporalStorage`.
        """
        pass

    @abstractmethod
    def values(self):
        """Returns the stored results iterator object. Each yield item is
        a ``(node_id, result)`` tuple."""

    @abstractmethod
    def keys(self):
        """Returns the stored tasks iterator object of the storage. Each
        yield item is a task ID."""

    def __contains__(self, task_id):
        """Checks if any of the task results are in the storage.
        Same as :meth:`PermanentStorage.contains(task_id)
        <PermanentStorage.contains>`."""
        return self.contains(task_id)

    def __delitem__(self, task_id):
        """Removes all the results from the storage. This is the same as
        :meth:`TemporalStorage.remove(task_id) <TemporalStorage.remove>`."""
        self.remove(task_id)

    def __iter__(self):
        """The same as :meth:`TemporalStorage.keys`."""
        return self.keys()

    def __len__(self):
        """The same as :meth:`TemporalStorage.count`."""
        return self.count


class PermanentStorage(object):
    """The interface for applications' permanent results storage.
    The storage can be a file, a database, a Python object in memory etc.

    The idea of the storage to store the tasks' results in the following
    fashion::

      {
          task_id1 : [res11, res12, ...],
          task_id2 : [res21, res22, ...],
          ...
      }

    Thus, :meth:`PermanentStorage.add(task_id, result) <PermanentStorage.add>`
    Initially adds the results to the storage, and appends the results if
    called the second, the third etc. time with the same ``task_id`` argument.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def add(self, task_id, result):
        """Stores the task result."""

    @abstractmethod
    def __getitem__(self, task_id):
        """Returns a list of task results.

        :rtype: :class:`list`
        """

    @abstractmethod
    def contains(self, task_id, result = None):
        """Checks if any of the task results or a particular task result
        is contained in the storage"""

    @abstractmethod
    def keys(self):
        """Returns the stored tasks iterator object of the storage. Each
        yield item is a task ID."""

    @abstractmethod
    def values(self):
        """Returns the stored results iterator object. Each yield item is
        a list of results associated with a particular task."""

    @abstractproperty
    def count(self):
        """The amount of unique task results in the storage.
        This is the same as:: ``len(list(ps.keys()))``, where ``ps`` is an
        instance of :class:`PermanentStorage`.
        """
        pass

    @abstractproperty
    def total_count(self):
        pass

    def __contains__(self, task_id):
        """Checks if any of the task results are in the storage.
        Same as :meth:`PermanentStorage.contains(task_id)
        <PermanentStorage.contains>`."""
        return self.contains(task_id)

    def __iter__(self):
        """The same as :meth:`PermanentStorage.keys`."""
        return self.keys()

    def __len__(self):
        """Same as :meth:`PermanentStorage.count`."""
        return self.count



# -*- coding: utf-8 -*-
import os
import sys
import stat
import tempfile
import shutil
import imp

from kaylee.testsuite import KayleeTest, load_tests
from kaylee.manager import (AdminCommandsManager, LocalCommandsManager,
                            BaseCommand)
from kaylee.util import nostdout

_pjoin = os.path.join

CURRENT_DIR = os.path.dirname(__file__)
RES_DIR = _pjoin(CURRENT_DIR, 'command_manager_tests_resources/')



def tmp_chdir():
    tmpdir = tempfile.mkdtemp(prefix='kl_unit_test__')
    os.chdir(tmpdir)
    return tmpdir


class KayleeCommandsManagerTests(KayleeTest):
    class CommandWithBlankName(BaseCommand):
        #pylint: disable-msg=W0223
        #W0223:  Method 'execute' is abstract in class 'BaseCommand'
        #        but is not overridden (throws NotImplementedError)
        name = ''

    def tearDown(self):
        tmpdir = tempfile.gettempdir()
        for tdir in os.listdir(tmpdir):
            if tdir.startswith('kl_unit_test__'):
                shutil.rmtree(os.path.join(tmpdir, tdir))

    def test_init(self):
        #pylint: disable-msg=R0201
        #R0201: Method could be a function
        AdminCommandsManager()

    def test_add_command(self):
        manager = LocalCommandsManager()
        self.assertRaises(ValueError,
                          manager.add_command,
                          self.CommandWithBlankName)

    def test_local_manager(self):
        manager = LocalCommandsManager()
        with nostdout():
            self.assertRaises(SystemExit, manager.parse, ['bad_command_name'])

    def test_admin_manager(self):
        manager = AdminCommandsManager()
        with nostdout():
            self.assertRaises(SystemExit, manager.parse, ['bad_command_name'])

    def test_start_env(self):
        manager = AdminCommandsManager()

        with nostdout():
            self.assertRaises(SystemExit, manager.parse, ['startenv'])

        tmpdir = tmp_chdir()
        with nostdout():
            manager.parse(['startenv', 'klenv'])

        # test whether files exist
        files_to_validate = [
            'klenv/klmanage.py',
            'klenv/settings.py',
        ]
        for fpath in files_to_validate:
            full_path = _pjoin(tmpdir, fpath)
            self.assertGreater(os.path.getsize(full_path), 0)

        self.assertTrue(os.stat(_pjoin(tmpdir, 'klenv/klmanage.py')).st_mode
                        & stat.S_IEXEC)
        # test settings contents
        settings_path = _pjoin(tmpdir, 'klenv/settings.py')
        settings = imp.load_source('tsettings', settings_path)
        self.assertEqual(settings.PROJECTS_DIR,
                         _pjoin(tmpdir, 'klenv'))

        # test 'startenv .' case
        tmpdir = tmp_chdir()
        with nostdout():
            manager.parse(['startenv', '.'])
        files_to_validate = [
            'klmanage.py',
            'settings.py',
        ]
        for fpath in files_to_validate:
            full_path = _pjoin(tmpdir, fpath)
            self.assertGreater(os.path.getsize(full_path), 0)


    def test_start_project(self):
        manager = LocalCommandsManager()
        with nostdout():
            self.assertRaises(SystemExit, manager.parse, ['startproject'])
            self.assertRaises(SystemExit, manager.parse,
                              ['startproject', 'ProjectName', '-m', 'x'])
            self.assertRaises(SystemExit, manager.parse,
                              ['startproject', 'ProjectName', '-t', 'abc'])

        # test for invalid project name
        self.assertRaises(ValueError, manager.parse, ['startproject', '@$'])
        self.assertRaises(ValueError, manager.parse,
                          ['startproject', 'Projec Name'])

        # test for generated project contents
        # create a project in a temporary current working dir
        tmpdir = tmp_chdir()

        with nostdout():
            manager.parse(['startproject', 'Monte_Carlo_Pi'])

        files_to_validate = [
            'monte_carlo_pi/client/monte_carlo_pi.js',
            'monte_carlo_pi/__init__.py',
            'monte_carlo_pi/monte_carlo_pi.py',
        ]

        self._validate_content(RES_DIR, tmpdir, files_to_validate)

    def test_start_project_coffee(self):
        manager = LocalCommandsManager()
        tmpdir = tmp_chdir()

        with nostdout():
            manager.parse(['startproject', 'Pi_Calc', '-t', 'coffee'])

        files_to_validate = [
            'pi_calc/client/pi_calc.coffee',
            'pi_calc/__init__.py',
            'pi_calc/pi_calc.py',
        ]

        self._validate_content(RES_DIR, tmpdir, files_to_validate)

    def test_build(self):
        lmanager = LocalCommandsManager()
        env_path = _start_env()
        os.chdir(env_path)

        # copy a test 'pi calc' project to the environment
        res_path = _pjoin(RES_DIR, 'pi_calc')
        dest_path = _pjoin(env_path, 'pi_calc')
        shutil.copytree(res_path, dest_path)

        # test plain is_kaylee_project_directory()
        sys.path.insert(0, env_path)
        from kaylee.manager.commands.build import is_kaylee_project_directory
        self.assertTrue(is_kaylee_project_directory(res_path))
        self.assertFalse(is_kaylee_project_directory('/etc'))
        del sys.path[0]

        with nostdout():
            lmanager.parse(['build'])

        build_path = os.path.join(env_path, '_build')
        self.assertTrue(os.path.exists(build_path), build_path)
        project_files_to_validate = [
            'js/pi_calc.js',
            'css/pi_calc.css',
            'css/other.css',
            'js/somelib.js',
            'js/otherlib.js',
            'data/somedata.dat',
            'data/otherdata',
        ]
        for fname in project_files_to_validate:
            fpath = os.path.join(build_path, 'pi_calc', fname)
            self.assertTrue(os.path.exists(fpath), fpath)

        kaylee_files_to_validate = [
            'js/kaylee.js',
            'js/klworker.js',
            'js/kldebug.js',
            'js/jquery.min.js',
            'css/klconsole.css'
        ]
        for fname in kaylee_files_to_validate:
            fpath = os.path.join(build_path, 'kaylee', fname)
            self.assertTrue(os.path.exists(fpath))

    def test_run(self):
        env_path = _start_env()
        os.chdir(env_path)
        lmanager = LocalCommandsManager()

        with nostdout():
            self.assertRaises(OSError, lmanager.parse, ['run'])


    def _validate_content(self, gtdir, tmpdir, files_to_validate):
        for fpath in files_to_validate:
            with open(_pjoin(tmpdir, fpath)) as f:
                generated_file_contents = f.read()
            with open(_pjoin(gtdir, fpath)) as f:
                ground_truth_file_contents = f.read().rstrip('\n')

            self.assertEqual(generated_file_contents,
                             ground_truth_file_contents)



def _start_env(name='tenv'):
    amanager = AdminCommandsManager()

    tmpdir = tmp_chdir()
    env_path = os.path.abspath(_pjoin(tmpdir, name))

    with nostdout():
        amanager.parse(['startenv', 'tenv'])
    return env_path





kaylee_suite = load_tests([KayleeCommandsManagerTests])



#pylint: disable-msg=W0212,W0611
#W0212: Access to a protected member
#W0611: Unused import PROJECTS_DIR # FALSE ALARM
###

from kaylee.testsuite import KayleeTest, load_tests, PROJECTS_DIR

import os
from kaylee import loader, Kaylee
from kaylee.contrib import (MemoryTemporalStorage,
                            MemoryPermanentStorage,
                            MemoryNodesRegistry)
from kaylee.session import JSONSessionDataManager
from kaylee.loader import Loader

_test_REGISTRY = {
    'name' : 'MemoryNodesRegistry',
    'config' : {
        'timeout' : '2s',
        },
}


class TestSettings(object):
    REGISTRY = _test_REGISTRY

    AUTO_GET_ACTION = True

    SECRET_KEY = '1234'

    SESSION_DATA_MANAGER = {
        'name' : 'JSONSessionDataManager',
        'config' : {}
    }


class TestSettingsWithApps(object):
    REGISTRY = _test_REGISTRY

    AUTO_GET_ACTION = True

    SECRET_KEY = '1234'

    APPLICATIONS = [
        {
            'name' : 'test.1',
            'description' : 'Test application',
            'project' : {
                'name' : 'AutoTestProject',
            },
            'controller' : {
                'name' :'TestController1',
                'config' : {},
                'temporal_storage' : {
                    'name' : 'MemoryTemporalStorage',
                },
                'permanent_storage' : {
                    'name' : 'MemoryPermanentStorage',
                },
            },
        }
    ]

    PROJECTS_DIR = PROJECTS_DIR


class KayleeLoaderTests(KayleeTest):
    def test_load_settings_class(self):
        kl = loader.load(TestSettings)
        self.assertIsInstance(kl, Kaylee)

    def test_load_settings_dict(self):
        # dict(Class.__dict__) wrapping: __dict__ is a dictproxy,
        kl = loader.load(dict(TestSettings.__dict__))
        self.assertIsInstance(kl, Kaylee)

    def test_load_settings_module(self):
        test_settings = __import__('test_settings')
        kl = loader.load(test_settings)
        self.assertIsInstance(kl, Kaylee)

    def test_load_settings_path(self):
        path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            'test_settings.py'))
        kl = loader.load(path)
        self.assertIsInstance(kl, Kaylee)

    def test_load_applications(self):
        settings = dict(TestSettingsWithApps.__dict__)
        ldr = Loader(settings)
        apps = ldr.applications
        self.assertIsInstance(apps, list)
        self.assertEqual(len(apps), 1)

        app = apps[0]
        self.assertEqual(app.__class__.__name__, 'TestController1')
        self.assertEqual(app.project.__class__.__name__, 'AutoTestProject')
        self.assertIsInstance(app.temporal_storage, MemoryTemporalStorage)
        self.assertIsInstance(app.permanent_storage, MemoryPermanentStorage)
        #self.assertIsInstance(app.project.storage, MemoryPermanentStorage)

    def test_load_registry(self):
        settings = dict(TestSettings.__dict__)
        ldr = Loader(settings)
        reg = ldr.registry
        self.assertIsInstance(reg, MemoryNodesRegistry)

    def test_load_session_data_manager(self):
        settings = dict(TestSettings.__dict__)
        ldr = Loader(settings)
        sdm = ldr.session_data_manager
        self.assertIsInstance(sdm, JSONSessionDataManager)

    def test_load_kaylee(self):
        kl = loader.load(TestSettingsWithApps)
        self.assertIsInstance(kl.registry, MemoryNodesRegistry)

        app = kl.applications['test.1']
        self.assertEqual(app.__class__.__name__, 'TestController1')
        self.assertEqual(app.project.__class__.__name__, 'AutoTestProject')

    def test_kaylee_setup(self):
        from kaylee import setup, kl
        self.assertIsNone(kl._wrapped)
        setup(TestSettingsWithApps)

        self.assertIsNotNone(kl._wrapped)
        self.assertIsInstance(kl.registry, MemoryNodesRegistry)

        app = kl.applications['test.1']
        self.assertEqual(app.__class__.__name__, 'TestController1')
        self.assertEqual(app.project.__class__.__name__, 'AutoTestProject')


kaylee_suite = load_tests([KayleeLoaderTests])



# -*- coding: utf-8 -*-
#pylint: disable-msg=W0703,W0212
#W0703: Catching too general exception Exception
#W0212: Access to a protected member a client class
#W0231: __init__ method from base class 'TestLoader' is not called

###
import os
import sys
import unittest
import logging
from importlib import import_module
from .helper import (KayleeTest, TestTemporalStorage, TestPermanentStorage,
                     TestController)

logging.basicConfig(level=logging.CRITICAL)
log = logging.getLogger(__name__)

unittest.defaultTestLoader = unittest.TestLoader()

PROJECTS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'projects')
sys.path.insert(0, PROJECTS_DIR)


def load_tests(test_cases):
    tsuite = unittest.TestSuite()
    for tcase in test_cases:
        loaded_suite = unittest.defaultTestLoader.loadTestsFromTestCase(tcase)
        tsuite.addTest(loaded_suite)
    return tsuite


def suite():
    """A testsuite that has all Kaylee tests.  You can use this
    function to integrate Kaylee tests into your own testsuite
    in case you want to test that monkeypatches to Kaylee do not
    break it.
    """
    tsuite = unittest.TestSuite()
    cdir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, cdir)
    for fname in os.listdir(cdir):
        if not fname.endswith('_tests.py'):
            continue
        modname = fname[:-3]
        try:
            mod = import_module(modname)
            if hasattr(mod, 'kaylee_suite'):
                tsuite.addTest(mod.kaylee_suite)
        except Exception as e:
            log.critical('Error importing module {}: {}'.format(modname, e))
            sys.exit(0)
    return tsuite


def find_all_tests(root_suite):
    """Yields all the tests and their names from a given suite."""
    suites = [root_suite]
    while suites:
        tsuite = suites.pop()
        try:
            # not that suite is iterable, thus every sub-suite from suite
            # is appended to the suites list
            suites.extend(tsuite)
        except TypeError:
            yield tsuite, '{}.{}.{}'.format(
                tsuite.__class__.__module__,
                tsuite.__class__.__name__,
                tsuite._testMethodName ).lower()


class KayleeTestsLoader(unittest.TestLoader):
    """
    A custom loader for Kaylee unit tests.
    Usage example:
    1. test.py kaylee_tests   # loads the suite from the module
    2. test.py KayleeLoadTest # loads particular TestCase
    3. test.py test_settings  # loads all(!) test functions with that name
    4. test.py KayleeLoadTest.test_settings # loads test function from TestCase
    5. test.py kaylee_tests.KayleeLoadTest, same as (2).
    """

    def __init__(self):
        unittest.TestLoader.__init__(self)
        self._default_suite = suite()

    def loadTestsFromName(self, name, module = None):
        name = name.lower()
        if name == 'default':
            return self._default_suite

        tests = []
        for testcase, testname in find_all_tests(self._default_suite):
            if (testname == name or
                testname.endswith('.' + name) or
                ('.' + name + '.') in testname or
                testname.startswith(name + '.') ):
                tests.append(testcase)

        if not tests:
            raise LookupError('Could not find test case for "{}"'.format(name))

        tsuite = unittest.TestSuite()
        if len(tests) == 1:
            return tests[0]
        for test in tests:
            tsuite.addTest(test)
        return tsuite


def main():
    """Runs the testsuite as command line application."""
    unittest.main(testLoader = KayleeTestsLoader(), defaultTest = 'default')



# -*- coding: utf-8 -*-
"""
    kaylee.node
    ~~~~~~~~~~~

    Implements Kaylee server-side node classes.
    The NodeID class is partially based on the code from pymongo package.
    See NOTICE for more details

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

import time
import binascii
import struct
import threading
import hashlib
from datetime import datetime
from abc import ABCMeta, abstractmethod

from .errors import (warn, InvalidNodeIDError, NodeNotSubscribedError,
                     ApplicationCompletedError)
from .util import parse_timedelta

#: The hex string formatted NodeID regular expression pattern which
#: can be used in e.g. web frameworks' URL dispatchers.
node_id_pattern = r'[\da-fA-F]{20}'


class Node(object):
    """
    A Node object contains the information about a registered `Kaylee Node`.

    :param node_id: an instance of :class:`NodeID` or a string parsable by
                    :class:`NodeID`
    """
    # __slots__ = ('id', '_task_id', 'subscription_timestamp', 'task_timestamp',
    #              'controller', 'errors_count', 'session_data')

    def __init__(self, node_id):
        if not isinstance(node_id, NodeID):
            raise TypeError('node_id must be an instance of {}, not {}'
                            .format(NodeID.__name__,  type(node_id).__name__ ))
        #: An instance of :class:`NodeID`
        self.id = node_id

        self.dirty = False
        self._subscription_timestamp = None
        self._task_timestamp = None
        self._controller = None
        self._session_data = None
        self._task_id = None

    def subscribe(self, controller):
        self._controller = controller
        self._subscription_timestamp = datetime.now()
        self.dirty = True
        return controller.project.client_config

    def unsubscribe(self):
        if self._controller is None:
            warn('Node.unsubscribe() is called for a non-subscribed node.')

        self._subscription_timestamp = None
        self._task_timestamp = None
        self._controller = None
        self._task_id = None
        #: Indicates that one of the Node attributes (except ID) has been
        #: changed. ``Node.dirty`` has to be set to ``False`` manually.
        self.dirty = True

    def get_task(self):
        if self.controller is None:
            raise NodeNotSubscribedError(self)
        if self.controller.completed:
            raise ApplicationCompletedError(self.controller)
        task = self.controller.get_task(self)
        task['id'] = str(task['id']).strip()
        return task

    def accept_result(self, result):
        if self.controller is None:
            raise NodeNotSubscribedError(self)
        if self.controller.completed:
            raise ApplicationCompletedError(self.controller)
        self.controller.accept_result(self, result)

    @property
    def controller(self):
        """Application which communicates with the node.
        It is an instance of :class:`Controller`."""
        return self._controller

    @property
    def session_data(self):
        """Binary session data (:class:`str`)"""
        return self._session_data

    @session_data.setter
    def session_data(self, val):
        self._session_data = val
        self.dirty = True

    @property
    def task_id(self):
        """The ID of the task being solved by the node."""
        return self._task_id

    @task_id.setter
    def task_id(self, val):
        self._task_id = val
        self._task_timestamp = datetime.now()
        self.dirty = True

    @property
    def subscription_timestamp(self):
        """A :class:`datetime.datetime` instance which tracks the time
        when a node has subscribed to an application."""
        return self._subscription_timestamp

    @property
    def task_timestamp(self):
        """A :class:`datetime.datetime` instance which tracks the time
        of a node receiving its last to-compute task."""
        return self._task_timestamp

    def __hash__(self):
        return hash(self.id)


class NodeID(object):
    """
    NodeID is a 10-bytes long ID generated from the current UNIX time,
    the remote host identifier and the internal incremental counter.
    The format is::

    [UNIX time (4)][counter (2)][remote host identifier hash (4)]

    Usually a new NodeID object is generated as follows::

        n = NodeID.for_host('192.168.10.20')

    The NodeID objects can be effectively compared in order to find out,
    which one was created earlier, e.g.::

        NodeID() < NodeID()
        # >>> True

    Is always ``True``, no matter what the remote host is.
    Another useful conversion is::

        n1 = NodeID()
        n2 = NodeID(str(NodeID))
        n1 == n2
        # >>> True

    The NodeID object can be also effectively used as a key in collections::

        n1 = NodeID()
        d = {n1: 'some_val'}
        n2 = NodeID(str(NodeID))
        n2 in d
        # >>> True

    :param node_id: A valid node id. The string representation of the
                    NodeID object is either binary 10-bytes long string
                    or 20-characters long hex string. If ``node_id``
                    is ``None``, then a new NodeID is generated.
    :param remote_host: Remote host IP address or other identifier.
    :type node_id: string, NodeID object or ``None``
    :type remote_host: string
    """
    #pylint: disable-msg=W0212
    #W0212: Access to a protected member _id of a client class (in __eg__ etc.)

    __slots__ = ('_id')
    _inc = 0
    _inc_lock = threading.Lock()

    def __init__(self, node_id=None, remote_host='127.0.0.1'):
        if node_id is None and not isinstance(remote_host, basestring):
            raise TypeError('remote_host must be an instance of {}, not {}'
                            .format(basestring.__name__,
                                    type(remote_host).__name__ )
                            )
        self._id = None
        if node_id is None:
            self._generate(remote_host)
        else:
            self._parse(node_id)

    @staticmethod
    def for_host(host):
        """Constructs a new NodeID object from remote host identifier.

        :param host: remote host identifier
        :type host: string
        :returns: NodeID object
        """
        return NodeID(remote_host=host)


    def _generate(self, remote_host):
        """Generates a new value for this NodeID."""
        nid = b''
        # 4 bytes current time
        nid += struct.pack('>i', int(time.time()))
        # 2 bytes inc
        with NodeID._inc_lock:
            nid += struct.pack(">i", NodeID._inc)[2:4]
            NodeID._inc = (NodeID._inc + 1) % 0xFFFF
        # 4 bytes host
        #pylint: disable-msg=E1101
        #E1101: Instance of 'md5' has no 'update' member
        host_hash = hashlib.md5()
        host_hash.update(remote_host)
        nid += host_hash.digest()[0:4]
        # 10 bytes total
        self._id = nid

    def _parse(self, nid):
        if isinstance(nid, NodeID):
            self._id = nid._id
        elif isinstance(nid, basestring):
            if len(nid) == 10:
                if isinstance(nid, str):
                    self._id = nid
                else:
                    raise InvalidNodeIDError(nid)
            elif len(nid) == 20:
                try:
                    self._id = nid.decode('hex')
                except (TypeError, ValueError):
                    raise InvalidNodeIDError(nid)
            else:
                raise InvalidNodeIDError(nid)
        else:
            raise TypeError('id must be an instance of {}, {} or {}, not {}'
                            .format(str.__name__,
                                    unicode.__name__,
                                    self.__class__.__name__,
                                    type(nid).__name__))

    @property
    def binary(self):
        """10-byte binary representation of the NodeID.

        :returns: string
        """
        return self._id

    @property
    def timestamp(self):
        """A :class:`datetime.datetime` instance representing
        the current NodeID's generation time. It is precise to a second.
        """
        t = struct.unpack(">i", self._id[0:4])[0]
        return datetime.fromtimestamp(t)

    def __str__(self):
        """Hex representation of the NodeID"""
        return binascii.hexlify(self._id).decode()

    def __repr__(self):
        return "NodeID('{}')".format(str(self))

    def __eq__(self, other):
        other = NodeID(other)
        return self._id == other._id

    def __ne__(self, other):
        other = NodeID(other)
        return self._id != other._id

    def __lt__(self, other):
        other = NodeID(other)
        return self._id < other._id

    def __le__(self, other):
        other = NodeID(other)
        return self._id <= other._id

    def __gt__(self, other):
        other = NodeID(other)
        return self._id > other._id

    def __ge__(self, other):
        other = NodeID(other)
        return self._id >= other._id

    def __hash__(self):
        """Python ``hash()`` of the internal id representation."""
        return hash(self._id)


class NodesRegistry(object):
    """
    The interface for registered nodes storage. A NodesRegistry is a place
    where Kaylee keeps information about active  Nodes. It can be as simple
    as Python collection or as complex as MongoDB or memcached - that is
    for the user to choose.

    :param timeout: Nodes timeout. Used by :meth:`clean` to determine if a
                    node is obsolete.
                    Format: ``1d 12h 59m 59s``, e.g.:

                    * ``'1d 10m'`` - one day, ten minutes
                    * ``'12h'``    - twelve hours
                    * ``'5h 10s'`` - five hours, ten seconds

    :type timeout: str
    """
    __metaclass__ = ABCMeta

    def __init__(self, timeout):
        #: Nodes timeout. Parsed from constructors ``timeout`` argument.
        #: Type: :class:`datetime.timedelta`.
        self.timeout = parse_timedelta(timeout)

    @abstractmethod
    def add(self, node):
        """Adds node to the storage."""

    @abstractmethod
    def update(self, node):
        """Updates previously added "dirty" nodes. There is no need to
        update a node if ``node.dirty == False``.

        :throws KeyError: in case node is not found in registry.
        """

    @abstractmethod
    def clean(self):
        """Removes the obsolete nodes from the storage."""

    @abstractmethod
    def __len__(self):
        """Returns the amount of nodes in the storage."""

    @abstractmethod
    def __delitem__(self, node):
        """Removes the node from the storage.

        :param node: an instance of :class:`Node` or a valid node id.
        """

    @abstractmethod
    def __getitem__(self, node_id):
        """Returns a node with the requested id."""

    @abstractmethod
    def __contains__(self, node):
        """Checks if the storage contains the node.

        :param node: an instance of :class:`Node` or a valid node id.
        """

def extract_node_id(node_or_node_id):
    """Extracts or constructs NodeID from the given object.

    :param node_or_node_id: a valid NodeID or previously initialized
                            Kaylee Node object.
    :type node_or_node_id: string, :class:`NodeID` or :class:`Node`
    :returns: :class:`NodeID` object
    """
    no = node_or_node_id
    if isinstance(no, basestring):
        return NodeID(node_id=no)
    elif isinstance(no, NodeID):
        return no
    elif isinstance(no, Node):
        return NodeID(node_id=no.id)
    else:
        raise TypeError('node must be an instance of {}, {}, or {} not'
                        ' {}'.format(basestring.__name__,
                                     NodeID.__name__,
                                     Node.__name__,
                                     type(no).__name__))



# -*- coding: utf-8 -*-
"""
    kaylee.project
    ~~~~~~~~~~~~~~

    This module provides basic interfaces for Kaylee projects.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

from abc import ABCMeta, abstractmethod


#: Defines auto project mode (see :attr:`Project.mode`)
AUTO_PROJECT_MODE = 0x2

#: Defines manual project mode (see :attr:`Project.mode`)
MANUAL_PROJECT_MODE = 0x4

#: Defines the unknown amount of tasks to be solved by a Kaylee
#: application.
UNKNOWN_AMOUNT = 0x1

KL_PROJECT_MODE = '__kl_project_mode__'
KL_PROJECT_SCRIPT_URL = '__kl_project_script_url__'
KL_PROJECT_STYLES = '__kl_project_styles__'



class Project(object):
    """Kaylee Projects abstract base class.

    :param script_url: The URL of the project's client part (\\*.js file).
    :param mode: defines :attr:`Project.mode <kaylee.Project.mode>`.
    """
    __metaclass__ = ABCMeta

    def __init__(self, script_url, mode, **kwargs):
        if mode not in [AUTO_PROJECT_MODE, MANUAL_PROJECT_MODE]:
            raise ValueError('Wrong project mode: {}'.format(mode))

        #: Indicates the mode in which project works on the client side.
        #: Available modes:
        #:
        #: * :data:`AUTO_PROJECT_MODE <kaylee.project.AUTO_PROJECT_MODE>`
        #: * :data:`MANUAL_PROJECT_MODE <kaylee.project.MANUAL_PROJECT_MODE>`
        #:
        #: For detailed description and usage see :ref:`projects_modes`.
        self.mode = mode

        #: A dictionary which contains the configuration passed to the
        #: client-side of the project in :js:func:`pj.init`.
        #: If the project is loaded from a :ref:`configuration object
        #: <loading>` the base value of ``client_config`` is extended by
        #: ``project.config`` configuration section.
        #:
        #: .. note:: Initially ``Project.client_config`` contains only the
        #:           data necessary to properly initialize the client-side
        #:           of the project.
        self.client_config = {
            KL_PROJECT_SCRIPT_URL: script_url,
            KL_PROJECT_MODE: self.mode,
            KL_PROJECT_STYLES: kwargs.get('styles', None),
        }

        #: Indicates whether the project was completed.
        self.completed = False

    @abstractmethod
    def next_task(self):
        """Returns the next task. The returned ``None`` value indicates that
        there will be no more new tasks from the project, but the bound
        controller can still refer to the old tasks via ``project[task_id]``.

        :returns: task :class:`dict` or ``None``.
        """

    @abstractmethod
    def __getitem__(self, task_id):
        """Returns a task with the required id. A task is simply
        a :class:`dict` with at least an 'id' key in it::

          {
              'id' : '10',
              'somedata' : somevalue,
              # etc.
          }

        :rtype: :class:`dict`
        """

    @abstractmethod
    def normalize_result(self, task_id, result):
        """Validates and normalizes the result.

        :param task_id: The ID of the task.
        :param result: The result to be validated and normalized.
        :throws ValueError: If the data is invalid.
        :return: normalized result.
        """

    def result_stored(self, task_id, data, storage):
        """A callback invoked by the bound controller when
        a result is successfully stored to a permanent storage.

        :param task_id: Task ID
        :param data: Normalized task result
        :param storage: The application's permanent results storage
        :type storage: :class:`PermanentStorage`
        """
        pass

    # @abstractproperty
    # def progress(self):
    #     """A tuple of 2 items indicating progress:
    #     ``(amount_of_tasks_completed, total_amount_of_tasks)``

    #     A negative (:data:`UNKNOWN_AMOUNT`) value of the second item in
    #     the tuple indicates that the project is not able to calculate the
    #     total amount of tasks.
    #     """
    #     pass



# -*- coding: utf-8 -*-

from werkzeug.routing import Map, Rule
from werkzeug.wrappers import Response
from kaylee import kl

def kaylee_register_node(request):
    reg_data = kl.register(request.remote_addr)
    return json_response(reg_data)

def kaylee_subscribe_node(request, app_name, node_id):
    #pylint: disable-msg=W0613
    #W0613:  Unused argument 'request'
    node_config = kl.subscribe(node_id, app_name)
    return json_response(node_config)

def kaylee_process_task(request, node_id):
    if request.method == 'GET':
        return json_response(kl.get_action(node_id))
    else:
        next_task = kl.accept_result(node_id, request.data)
        # the reason for using request.data instead of request.json
        # is that Kaylee expects the "raw", non-processed data
        return json_response(next_task)

def json_response(s):
    return Response(s, mimetype = 'application/json')

def make_url_map(url_prefix='/kaylee'):
    return Map([
        Rule(url_prefix + '/register',
             methods=['GET'],
             endpoint=kaylee_register_node),
        Rule(url_prefix + '/apps/<app_name>/subscribe/<node_id>',
             methods=['POST'],
             endpoint=kaylee_subscribe_node),
        Rule(url_prefix + '/actions/<node_id>',
             methods=['GET', 'POST'],
             endpoint=kaylee_process_task)
    ])



import os
import stat
from jinja2 import Template
from kaylee.manager import AdminCommand
from kaylee.util import random_string


class StartEnvCommand(AdminCommand):
    name = 'startenv'
    help = 'Creates Kaylee environment.'

    args = {
        'name' : {},
    }

    @staticmethod
    def execute(opts):
        start_env(opts)

#pylint: disable-msg=R0914
#R0914: 21,0:start_env: Too many local variables (16/15)
def start_env(opts):
    ENV_TEMPLATE_DIR = 'templates/env_template'
    ENV_TEMPLATE_PATH = os.path.join(os.path.dirname(__file__),
                                     ENV_TEMPLATE_DIR)
    TEMPLATE_FILES = [
        # see notes about klmanage.py below
        ('klmanage.py', 'klmanage.py'),
        ('settings.py', 'settings.py'),
    ]

    if opts.name == '.':
        dest_path = os.getcwd()
    else:
        dest_path = os.path.join(os.getcwd(), opts.name)
        os.mkdir(dest_path)

    render_args = {
        'SECRET_KEY' : random_string(32),
        'PROJECTS_DIR' : dest_path,
    }

    for fname, out_fname in TEMPLATE_FILES:
        template_path = os.path.join(ENV_TEMPLATE_PATH, fname)
        with open(template_path) as f:
            template_data = f.read()
        document_data = Template(template_data).render(**render_args)
        out_path = os.path.join(dest_path, out_fname)
        with open(out_path, 'w') as f:
            f.write(document_data)

    # explicitly chmod +x klmanage.py
    klmanage_path = os.path.join(dest_path, 'klmanage.py')
    st = os.stat(klmanage_path)
    os.chmod(klmanage_path, st.st_mode | stat.S_IEXEC)

    _dirname = 'current directory' if opts.name == '.' else opts.name
    print('Kaylee environment has been successfully created in {}'
          .format(_dirname))



# -*- coding: utf-8 -*-
import os

from werkzeug.wrappers import Request, Response
from werkzeug.exceptions import HTTPException
from werkzeug.wsgi import SharedDataMiddleware
from werkzeug.routing import Rule
from werkzeug.serving import run_simple

import kaylee
from kaylee.contrib.frontends.werkzeug_frontend import make_url_map
from kaylee.util import setup_logging

import logging
log = logging.getLogger(__name__)

url_map = make_url_map(url_prefix='/kaylee')

@Request.application
def application(request):
    adapter = url_map.bind_to_environ(request.environ)
    try:
        endpoint, values = adapter.match()
        return endpoint(request, **values)
    except HTTPException as e:
        return e


def run(settings_file, static_dir, port=5000, debug=False):
    loglevel = logging.DEBUG if debug else logging.INFO
    setup_logging(loglevel)
    kaylee.setup(settings_file)

    # index.html at 'http://server.address/' URL
    with open(os.path.join(static_dir, 'index.html')) as f:
        index_data = f.read()
    #pylint: disable-msg=W0613
    #W0613 unused 'request'
    def _home(request):
        return Response(index_data, mimetype='text/html')
    home_rule = Rule('/', methods=['GET'], endpoint=_home)
    url_map.add(home_rule)

    log.debug(static_dir)
    # add static data middleware and start the server
    app = SharedDataMiddleware(application,
                               {'/static': static_dir })

    run_simple('127.0.0.1', port, app, use_debugger=True,
               use_reloader=False)




# -*- coding: utf-8 -*-
"""
    kaylee.loader
    ~~~~~~~~~~~~~

    Implements projects, controllers and Kaylee instance loader.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

import os
import imp
import importlib
import inspect
import types
from copy import deepcopy
from collections import defaultdict

import kaylee.contrib
from .core import Kaylee
from .errors import KayleeError
from .util import LazyObject, is_strong_subclass
from . import storage, controller, project, node, session

import logging
log = logging.getLogger(__name__)

_default_settings = {
    'AUTO_GET_ACTION' : True,
}


class LazyKaylee(LazyObject):
    def _setup(self, obj = None):
        if isinstance(obj, Kaylee):
            self._wrapped = obj
        elif obj is None:
            self._wrapped = None # release the object
        else:
            raise TypeError('obj must be an instance of {} or '
                            'a Kaylee settings object, not {}'
                            .format(Kaylee.__name__, type(obj).__name__))


def load(settings):
    """Loads Kaylee.

    :param settings: Kaylee settings object.
    :type settings: dict, class, module or absolute Python module path.
    :returns: Kaylee object.
    """
    # Whatever the initial settings type is, load it (optional) and
    # convert it to a dictionary.
    # Note, that only settings with uppercase names' are loaded.
    if isinstance(settings, basestring):
        if os.path.exists(settings):
            settings = imp.load_source('kl_settings', settings)
    if isinstance(settings, (type, types.ModuleType)):
        d = {}
        for attr in dir(settings):
            if attr == attr.upper():
                d[attr] = getattr(settings, attr)
        settings = d
    if isinstance(settings, dict):
        settings = {k : v for k, v in settings.iteritems() if k == k.upper()}

    # at this point, if settings is not a dict, then object type is wrong.
    if not isinstance(settings, dict):
        raise TypeError('settings must be an instance of '
                        '{}, {}, {} or {} not {}'.format(
                            dict.__name__,
                            type.__name__,
                            types.ModuleType.__name__,
                            basestring.__name__,
                            type(settings).__name__))

    new_settings = deepcopy(_default_settings)
    new_settings.update(settings)
    try:
        loader = Loader(new_settings)
        registry = loader.registry
        sdm = loader.session_data_manager
        apps = loader.applications
    except (KeyError, AttributeError) as e:
        raise KayleeError('Settings error or object was not found: "{}"'
                          .format(e.args[0]))
    return Kaylee(registry=registry,
                  session_data_manager=sdm,
                  applications=apps,
                  **settings)



class Loader(object):
    _loadable_base_classes = [
        project.Project,
        controller.Controller,
        node.NodesRegistry,
        storage.PermanentStorage,
        storage.TemporalStorage,
        session.SessionDataManager,
    ]

    def __init__(self, settings):
        self._classes = defaultdict(dict)
        self._settings = settings

        # load classes from contrib (non-refreshable)
        self._update_classes(kaylee.contrib)
        # load session data managers
        self._update_classes(kaylee.session)
        # load classes from project modules (refreshable for new modules only)
        if 'PROJECTS_DIR' in settings:
            projects_dir = settings['PROJECTS_DIR']
            for mod in find_modules(projects_dir):
                self._update_classes(mod)

    @property
    def registry(self):
        settings = self._settings
        clsname = settings['REGISTRY']['name']
        regcls = self._classes[node.NodesRegistry][clsname]
        return regcls(**settings['REGISTRY']['config'])

    @property
    def session_data_manager(self):
        settings = self._settings
        if 'SESSION_DATA_MANAGER' in settings:
            clsname = settings['SESSION_DATA_MANAGER']['name']
            sdmcls = self._classes[session.SessionDataManager][clsname]
            sdm_config = settings['SESSION_DATA_MANAGER'].get('config', {})
        else:
            # Load default (should be Phony) session data manager in case it
            # is not defined in settings.
            default_manager = session.KL_LOADER_DEFAULT_SESSION_DATA_MANAGER
            sdmcls = default_manager
            sdm_config = {}
        return sdmcls(**sdm_config)

    @property
    def applications(self):
        settings = self._settings
        apps = []
        if 'APPLICATIONS' in settings:
            for conf in settings['APPLICATIONS']:
                ct = self._load_controller(conf)
                apps.append(ct)
        return apps

    def _update_classes(self, module):
        """Updates the _classes field by the classes found in
        the module."""
        classes_from_module = get_classes_from_module(module)
        for base_class in self._loadable_base_classes:
            # update _classes[base_class] with
            # { class_name : class } pairs, where `class` is a subclass of
            # `base_class`
            name_class_pairs = {c.__name__ : c for c in classes_from_module
                                if is_strong_subclass(c, base_class)}
            self._classes[base_class].update(name_class_pairs)

    def _load_permanent_storage(self, conf):
        psconf = conf['controller']['permanent_storage']
        clsname = psconf['name']
        pscls = self._classes[storage.PermanentStorage][clsname]
        return pscls(**psconf.get('config', {}))

    def _load_temporal_storage(self, conf):
        if not 'temporal_storage' in conf['controller']:
            return None
        tsconf = conf['controller']['temporal_storage']
        clsname = tsconf['name']
        tscls = self._classes[storage.TemporalStorage][clsname]
        return tscls(**tsconf.get('config', {}))

    def _load_project(self, conf):
        clsname = conf['project']['name']
        pcls = self._classes[project.Project][clsname]
        pj_config = conf['project'].get('config', {})
        return pcls(**pj_config)

    def _load_controller(self, conf):
        # initialize objects
        clsname = conf['controller']['name']
        ccls = self._classes[controller.Controller][clsname]
        app_name = conf['name']
        pjobj = self._load_project(conf)
        psobj = self._load_permanent_storage(conf)
        tsobj = self._load_temporal_storage(conf)

        cobj = ccls(app_name, pjobj, psobj, tsobj,
                    **conf['controller'].get('config', {}))
        return cobj


def get_classes_from_module(module):
    return [attr for attr in module.__dict__.values()
            if inspect.isclass(attr)]

def find_packages(path):
    for sub_dir in os.listdir(path):
        pdir_path = os.path.join(path, sub_dir)
        if not os.path.isdir(pdir_path):
            continue
        if '__init__.py' not in os.listdir(pdir_path):
            continue
        yield sub_dir

def find_modules(path):
    """A generator which yields python modules found in the given path."""
    for package_name in find_packages(path):
        try:
            pymod = importlib.import_module(package_name)
            yield pymod
        except ImportError as e:
            log.error('Unable to import project package "{}": {}'
                      .format(package_name, e))



# -*- coding: utf-8 -*-
"""
    kaylee.contrib.registries
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    The module provides basic Kaylee Node registry implementation(s).

    :copyright: (c) 2013 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

from datetime import datetime
from kaylee.node import NodesRegistry, NodeID, extract_node_id


class MemoryNodesRegistry(NodesRegistry):
    def __init__(self, *args, **kwargs):
        super(MemoryNodesRegistry, self).__init__(*args, **kwargs)
        self._d = {}

    def add(self, node):
        self._d[node.id] = node

    def update(self, node):
        # a very naive and simple update
        if node.id in self._d and node.dirty:
            self._d[node.id] = node
        else:
            raise KeyError('Cannot update node in registry: '
                           'node {} was not found'.format(node))

    def clean(self):
        nodes_to_clean = (node for node in self._d.iteritems()
                          if datetime.now() - node.id.timestamp > self.timeout)
        for node in nodes_to_clean:
            del self._d[node]

    def __len__(self):
        return len(self._d)

    def __delitem__(self, node):
        node_id = extract_node_id(node)
        try:
            del self._d[node_id]
        except KeyError:
            pass

    def __getitem__(self, node_id):
        node_id = extract_node_id(node_id)
        return self._d[node_id]

    def __contains__(self, node):
        node_id = extract_node_id(node)
        return node_id in self._d




# -*- coding: utf-8 -*-
import math

from kaylee.testsuite import load_tests
from kaylee.testsuite.helper import SubclassTestsBase
from kaylee.storage import TemporalStorage, PermanentStorage
from kaylee.contrib.storages import (MemoryTemporalStorage,
                                     MemoryPermanentStorage)
from copy import deepcopy
from kaylee.node import NodeID


def _tgen(i, prefix='t'):
    return "{}{}".format(prefix, i)

def _rgen(i, prefix='r'):
    return "{}{}".format(prefix, i)


class TemporalStorageTestsBase(SubclassTestsBase):
    def setUp(self):
        super(TemporalStorageTestsBase, self).setUp()

    def test_init(self):
        ts = self.cls_instance()
        self.assertIsInstance(ts, TemporalStorage)
        self.assertEqual(len(ts), 0)

    def test_add(self):
        # add one
        ts = self.cls_instance()
        node_id = NodeID()
        ts.add('t1', node_id, 'r1')
        self.assertEqual(list(ts['t1']), [(node_id, 'r1'), ])

        # add many
        ts = self.cls_instance()
        node_id = NodeID()
        self._fill_storage(ts, self.MANY, node_id=node_id)
        for i in range(0, self.MANY):
            tid = _tgen(i)
            for nid, res in ts[tid]:
                self.assertEqual(nid, node_id)
                self.assertEqual(res, _rgen(i))

    def test_add_complex_result(self):
        ts = self.cls_instance()
        test_res = {
            'edx' : [1, 2, 3],
            'eax' : { (3, 4) : (5, 6), 'c' : 10},
            1 : 100,
            2 : {
                3 : 'abc',
                'cde' : [5, 6, 7] * 100,
                0 : 'c' * 65537
            }
        }
        ts.add('1', NodeID(), deepcopy(test_res))
        nr_list = list(ts['1'])
        added_res = nr_list[0][1]
        self.assertEqual(test_res, added_res)

    def test_overwrite(self):
        # overwrite one
        ts = self.cls_instance()
        node_id = NodeID()
        ts.add('t1', node_id, 'r1')
        ts.add('t1', node_id, 'r2')
        self.assertEqual(list(ts['t1']), [(node_id, 'r2'), ])

        # overwrite many
        ts = self.cls_instance()
        rgen = 'XYZ{}'.format
        self._fill_storage(ts, self.SOME, node_id=node_id)
        self._fill_storage(ts, self.SOME, rgen_func=rgen, node_id=node_id)

        for i in range(0, self.SOME):
            tid = _tgen(i)
            for nid, res in ts[tid]:
                self.assertEqual(nid, node_id)
                self.assertEqual(res, rgen(i))

    def test_remove_and_del(self):
        # remove one
        ts = self.cls_instance()
        node_id = NodeID()
        ts.add('t1', node_id, 'r1')
        ts.remove('t1')
        self.assertRaises(KeyError, ts.__getitem__, 't1')
        self.assertEqual(len(ts), 0)

        # add one, remove two
        ts.add('t1', node_id, 'r1')
        self.assertRaises(KeyError, ts.remove, 't1', NodeID())
        self.assertRaises(KeyError, ts.remove, 't2')
        ts.remove('t1', node_id)
        self.assertEqual(list(ts['t1']), [])
        self.assertEqual(len(ts), 1)
        ts.remove('t1')
        self.assertEqual(len(ts), 0)
        self.assertRaises(KeyError, ts.__getitem__, 't1')

        # test del
        ts.add('t3', node_id, 'r1')
        del ts['t3']
        self.assertRaises(KeyError, ts.__getitem__, 't3')
        self.assertEqual(len(ts), 0)

        # remove many
        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME, node_id=node_id)
        self.assertEqual(len(ts), self.SOME)
        for i in range(0, self.SOME):
            tid_res = ts[_tgen(i)]
            del ts[_tgen(i)]
            self.assertRaises(KeyError, ts.__getitem__, list(tid_res)[0])
        self.assertEqual(len(ts), 0)

    def test_clear(self):
        # fill and clear
        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME)
        self.assertEqual(len(ts), self.SOME)
        ts.clear()
        self.assertEqual(len(ts), 0)

        # double fill and clear
        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME)
        self._fill_storage(ts, self.SOME)
        self.assertEqual(len(ts), self.SOME)
        ts.clear()
        self.assertEqual(len(ts), 0)

        # same ts, fill and clear
        self._fill_storage(ts, self.SOME, node_id=NodeID())
        self.assertEqual(len(ts), self.SOME)
        ts.clear()
        self.assertEqual(len(ts), 0)

        # remove all and clear
        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME, node_id=NodeID())
        for i in range(0, self.SOME):
            ts.remove(_tgen(i))
        self.assertEqual(len(ts), 0)
        ts.clear()
        self.assertEqual(len(ts), 0)

    def test_contains_and_in(self):
        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME)

        # by task id
        for i in range(0, self.SOME):
            tid = _tgen(i)
            self.assertTrue(ts.contains(tid))
            self.assertIn(tid, ts)

        self.assertFalse(ts.contains('tx'))
        self.assertNotIn('tx', ts)

        # by node id
        ts = self.cls_instance()
        node_id = NodeID()
        self._fill_storage(ts, self.SOME, node_id=node_id)
        for i in range(0, self.SOME):
            tid = _tgen(i)
            self.assertTrue(ts.contains(tid, node_id))
            self.assertFalse(ts.contains(tid, NodeID()))
            self.assertFalse(ts.contains(_tgen(i + self.SOME), node_id))
            self.assertIn(tid, ts)

        # after clear()
        ts.clear()
        for i in range(0, self.SOME):
            tid = _tgen(i)
            self.assertFalse(ts.contains(tid))
            self.assertFalse(ts.contains(tid, node_id))
            self.assertNotIn(tid, ts)

    def test_count_and_total_count(self):
        ts = self.cls_instance()
        # test simple amount of stored values
        self._fill_storage(ts, self.SOME)
        self.assertEqual(len(ts), self.SOME)
        self.assertEqual(ts.count, self.SOME)
        self.assertEqual(ts.total_count, self.SOME)

        # double fill
        self._fill_storage(ts, self.SOME)
        self._fill_storage(ts, self.SOME)
        self.assertEqual(len(ts), self.SOME)
        self.assertEqual(ts.count, self.SOME)
        self.assertEqual(ts.total_count, 3 * self.SOME)

        # remove
        ts.remove('t0')
        self.assertEqual(len(ts), self.SOME - 1)
        self.assertEqual(ts.count, self.SOME - 1)
        self.assertEqual(ts.total_count, 3 * self.SOME - 3)

        # remove with specific node id
        nid = NodeID()
        ts.add('t1', nid, 'r12')
        self.assertEqual(len(ts), self.SOME - 1)
        self.assertEqual(ts.count, self.SOME - 1)
        self.assertEqual(ts.total_count, 3 * self.SOME - 2)

        ts.remove('t1', nid)
        self.assertEqual(len(ts), self.SOME - 1)
        self.assertEqual(ts.count, self.SOME - 1)
        self.assertEqual(ts.total_count, 3 * self.SOME - 3)

        # add one, remove
        ts.add('t0', NodeID(), 'r0')
        for i in range(0, self.SOME):
            ts.remove(_tgen(i))
        self.assertEqual(len(ts), 0)
        self.assertEqual(ts.count, 0)
        self.assertEqual(ts.total_count, 0)

        # more double fill
        self._fill_storage(ts, self.MANY)
        self._fill_storage(ts, self.MANY)
        self.assertEqual(len(ts), self.MANY)
        self.assertEqual(ts.count, self.MANY)
        self.assertEqual(ts.total_count, 2 * self.MANY)

        # test nodes count
        nodes = set(n for n, r in ts.values())
        self.assertEqual(len(nodes), 2 * self.MANY)

    def test_keys_and_iter(self):
        def _basic_ki_test(ts, count):
            test_task_ids = set(_tgen(i) for i in range(0, count))
            iter_task_ids = set(tid for tid in iter(ts))
            keys_task_ids = set(tid for tid in ts.keys())
            self.assertEqual(test_task_ids, iter_task_ids)
            self.assertEqual(test_task_ids, keys_task_ids)
            self.assertEqual(len(iter_task_ids), len(list(iter(ts))))
            self.assertEqual(len(keys_task_ids), len(list(ts.keys())))

        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME)
        _basic_ki_test(ts, self.SOME)

        # fill more, same test
        self._fill_storage(ts, self.SOME)
        _basic_ki_test(ts, self.SOME)

        # remove self.SOME, same test
        for i in range(0, self.SOME):
            ts.remove(_tgen(i))
        _basic_ki_test(ts, 0)

        # clear, test with the same NodeID
        ts.clear()
        node_id = NodeID()
        self._fill_storage(ts, self.SOME, node_id=node_id)
        self._fill_storage(ts, self.SOME)

        for i in range(0, self.SOME):
            ts.remove(_tgen(i), node_id)
        _basic_ki_test(ts, self.SOME)

    def test_values(self):
        ts = self.cls_instance()
        self._fill_storage(ts, self.SOME)
        self.assertEqual(len(list(ts.values())), self.SOME)

        test_results_set = set(_rgen(i) for i in range(0, self.SOME))
        for nr in ts.values():
            self.assertIn(nr[1], test_results_set)
            self.assertEqual(NodeID(nr[0]).binary, nr[0])

        # fill once more and count
        self._fill_storage(ts, self.SOME)
        for nr in ts.values():
            self.assertIn(nr[1], test_results_set)
            self.assertEqual(NodeID(nr[0]).binary, nr[0])
        self.assertEqual(len(list(ts.values())), 2 * self.SOME)

        # clear and count
        ts.clear()
        self.assertEqual(len(list(ts.values())), 0)

        # specify node id
        node_id = NodeID()
        self._fill_storage(ts, self.SOME, node_id=node_id)
        for nr in ts.values():
            self.assertIn(nr[1], test_results_set)
            self.assertEqual(NodeID(nr[0]).binary, node_id.binary)

        # remove a result by task_id
        HALF_SOME = self.SOME // 2
        for i in range(0, HALF_SOME):
            ts.remove(_tgen(i))
        for nr in ts.values():
            res = nr[1]
            res_id = res.split('r')[1]
            self.assertGreaterEqual(int(res_id), HALF_SOME)

        # fill more and count
        self._fill_storage(ts, self.SOME)
        # math.ceil because range(0, HALF_SOME)
        expected_count = self.SOME + math.ceil(self.SOME / 2.0)
        self.assertEqual(len(list(ts.values())), expected_count)

    @staticmethod
    def _fill_storage(ts, count, tgen_func=_tgen, rgen_func=_rgen,
                      node_id=None):
        if node_id is None:
            # if node_id has not been defined by a user
            # define callable _nodeid == NodeID
            _nodeid = NodeID
        else:
            # a callable which returns the `node_id` argument
            _nodeid = lambda: node_id
        for i in range(0, count):
            ts.add(tgen_func(i), _nodeid(), rgen_func(i))


class PermanentStorageTestsBase(SubclassTestsBase):
    def setUp(self):
        super(PermanentStorageTestsBase, self).setUp()

    def test_init(self):
        ps = self.cls_instance()
        self.assertIsInstance(ps, PermanentStorage)
        self.assertEqual(len(ps), 0)

    def test_add(self):
        ps = self.cls_instance()
        ps.add('1', 'r1')

        self.assertEqual(ps['1'], ['r1'])
        self.assertEqual(len(ps), 1)
        self.assertEqual(list(iter(ps)), ['1'])

        # add many
        ps = self.cls_instance()
        self._fill_storage(ps, self.MANY)
        self.assertEqual(len(ps), self.MANY)

        for i in range(0, self.MANY):
            task_id = _tgen(i)
            result = [_rgen(i)]
            self.assertEqual(ps[task_id], result)

    def test_add_same_task_results(self):
        ps = self.cls_instance()
        self._fill_storage(ps, self.SOME)
        self._fill_storage(ps, self.SOME)
        self._fill_storage(ps, self.SOME)

        for i in range(0, self.SOME):
            task_id = _tgen(i)
            result = [_rgen(i)] * 3
            self.assertEqual(ps[task_id], result)

    def test_add_complex_result(self):
        ps = self.cls_instance()
        res = {
            'edx' : [1, 2, 3],
            'eax' : { (3, 4) : (5, 6), 'c' : 10},
            1 : 100,
            2 : {
                3 : 'abc',
                'cde' : [5, 6, 7] * 100,
                0 : 'c' * 65537
            }
        }
        ps.add('1', deepcopy(res))
        self.assertEqual(ps['1'], [res, ])

        ps.add('1', deepcopy(res))
        self.assertEqual(ps['1'], [res, res])

        for i in range(0, self.SOME):
            ps.add(_tgen(i), deepcopy(res))
            self.assertEqual(ps[_tgen(i)], [res])

    def test_keys_iter(self):
        ps = self.cls_instance()
        self._fill_storage(ps, self.SOME)
        test_keys_set = set(_tgen(i) for i in range(0, self.SOME))
        ps_iter_keys_set = set(iter(ps))
        self.assertEqual(len(list(iter(ps))), self.SOME)
        self.assertEqual(test_keys_set, ps_iter_keys_set)
        self.assertEqual(list(iter(ps)), list(ps.keys()))

    def test_values_iter(self):
        ps = self.cls_instance()
        self._fill_storage(ps, self.SOME)
        test_values_list = [[_rgen(i)] for i in range(0, self.SOME)]
        self.assertEqual(sorted(test_values_list), sorted(list(ps.values())) )

    def test_count_and_total_count(self):
        ps = self.cls_instance()
        self._fill_storage(ps, self.SOME)
        self.assertEqual(ps.count, len(ps))
        self.assertEqual(ps.count, self.SOME)
        self.assertEqual(ps.total_count, self.SOME)

        ps.add('t0', 'r0')
        self._fill_storage(ps, self.SOME)
        self.assertEqual(ps.count, self.SOME)
        self.assertEqual(ps.total_count, self.SOME * 2 + 1)

    def test_contains(self):
        ps = self.cls_instance()
        self._fill_storage(ps, self.SOME)
        self.assertTrue(ps.contains('t0'))
        self.assertTrue(ps.contains("t{}".format(self.SOME - 1)))
        self.assertIn('t0', ps)
        self.assertIn("t{}".format(self.SOME - 1), ps)
        self.assertIn("t{}".format(self.SOME // 2), ps)
        self.assertFalse(ps.contains('xx'))
        self.assertNotIn('xx', ps)
        self.assertTrue(ps.contains('t0', 'r0'))
        self.assertTrue(ps.contains("t{}".format(self.SOME - 1),
                                    "r{}".format(self.SOME - 1)) )
        self.assertFalse(ps.contains('t0', 'aa'))

        ps.add('t0', 'abc')
        self.assertTrue(ps.contains('t0', 'abc'))
        self.assertTrue(ps.contains('t0', 'r0'))

        for i in range(0, self.MANY):
            res = "{}{}".format('n', i)
            ps.add('t0', res)
            self.assertTrue(ps.contains('t0', res))

    @staticmethod
    def _fill_storage(ps, count, tgen_func=_tgen, rgen_func=_rgen):
        for i in range(0, count):
            ps.add(tgen_func(i), rgen_func(i))


class MemoryPermanentStorageTests(PermanentStorageTestsBase):
    def cls_instance(self):
        return MemoryPermanentStorage()


class MemoryTemporalStorageTests(TemporalStorageTestsBase):
    def cls_instance(self):
        return MemoryTemporalStorage()


kaylee_suite = load_tests([
   MemoryTemporalStorageTests,
   MemoryPermanentStorageTests
])



# -*- coding: utf-8 -*-
"""
    kaylee.contrib
    ~~~~~~~~~~~~~~

    This sub-package contains code contributed to Kaylee.
    It includes front-ends, Controllers, Storages, NodeRegistries etc.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

from .controllers import SimpleController, ResultsComparatorController
from .storages import MemoryTemporalStorage, MemoryPermanentStorage
from .registries import MemoryNodesRegistry



import unittest
from kaylee.project import Project
from kaylee.contrib.storages import (MemoryTemporalStorage,
                                     MemoryPermanentStorage)
from kaylee.contrib.controllers import SimpleController
from abc import ABCMeta, abstractmethod


class KayleeTest(unittest.TestCase):
    """Base class for all Kaylee tests."""

class SubclassTestsBase(KayleeTest):
    """The base class for (sub)classes, e.g. controllers, storages etc.
    tests. The idea that Kaylee provides basic tests, which can then be
    subclassed and extentended by class-specific tests"""
    SOME = 11
    MANY = SOME ** 2

    __metaclass__ = ABCMeta

    def setUp(self):
        self.cls = self.cls_instance().__class__

    #pylint: disable-msg=R0201
    #R0201: Method could be a function
    @abstractmethod
    def cls_instance(self):
        """Returns an instance of """
        return None


class NonAbstractProject(Project):
    def __init__(self, *args, **kwargs):
        super(NonAbstractProject, self).__init__(*args, **kwargs)

    def __getitem__(self, task_id):
        pass

    def next_task(self):
        pass

    def normalize_result(self, task_id, result):
        pass

    def result_stored(self, task_id, result, storage):
        pass


class TestTemporalStorage(MemoryTemporalStorage):
    pass


class TestPermanentStorage(MemoryPermanentStorage):
    pass


class TestController(SimpleController):
    @staticmethod
    def new_test_instance(project_object=None, app_name='test_app'):
        # Load autoproject from testsuite in order to be able to
        # completely initialize a test controller with all necessary
        # attributes.
        if project_object is None:
            from kaylee.testsuite.projects.auto_test_project \
                import AutoTestProject
            project_object = AutoTestProject()

        return TestController(app_name,
                              project_object,
                              TestPermanentStorage(),
                              TestTemporalStorage())



# -*- coding: utf-8 --*
"""
    kaylee
    ~~~~~~

    Kaylee is distributed and volunteer computing framework.

    It implements all the functionality required to distribute computation
    among nodes and collect the results. Kaylee aims to let a user
    completely focus on project's business logic and handle all other
    stuff automatically.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

__version__ = '0.3'

from . import loader
kl = loader.LazyKaylee()

from .core import Kaylee, Applications
from .node import Node, NodeID, NodesRegistry
from .storage import (TemporalStorage,
                      PermanentStorage )
from .session import SessionDataManager
from .controller import Controller
from .project import Project
from .errors import (KayleeError,
                     InvalidNodeIDError,
                     NodeNotSubscribedError,
                     InvalidResultError,
                     NodeRequestRejectedError,
                     ApplicationCompletedError,)


def setup(settings):
    #pylint: disable-msg=W0212
    #W0212: Access to a protected member _setup
    if settings is not None:
        # In this case we are trying to load Kaylee from the given settings
        kl._setup(loader.load(settings))
    else:
        # In this case the user asks us to "release" Kaylee
        # instance from the proxy.
        kl._setup(None)



import os
import re
import shutil
from jinja2 import Template
from kaylee.manager import LocalCommand

# TODO: check other commands for shutil.copytree

class StartProjectCommand(LocalCommand):
    name = 'startproject'
    help = 'Starts new Kaylee project'

    args = {
        'name' : {},
        ('-m', '--mode') : dict(
            choices=['manual', 'auto'],
            default='auto',
            help="project mode",
        ),
        ('-t', '--template') : dict(
            choices=['js', 'coffee'],
            default='js',
            help="client-side programming language.",
        ),
    }

    @staticmethod
    def execute(opts):
        validate_name(opts)
        start_project(opts)


def validate_name(opts):
    if re.match(r'^\w+$', opts.name) is None:
        raise ValueError('Invalid project name: {} ([A-Za-z0-9_])'
                         .format(opts.name))


def start_project(opts):
    PROJECT_TEMPLATE_DIR = 'templates/project_template'
    PROJECT_TEMPLATE_PATH = os.path.join(os.path.dirname(__file__),
                                         PROJECT_TEMPLATE_DIR)
    TEMPLATE_FILES = [
        #(template file in PROJECT_TEMPLATE_DIR,
        # destination file name with {project} macro replacement)
        # e.g. ('project.py', '{project_name}.py'),
        ('__init__.py.template', '__init__.py'),
        ('project.py.template', '{project_name}.py'),
        _client_template_files(opts)
    ]

    # build rendering environment constants
    project_file_name = opts.name.lower()

    # copy project template to cwd
    dest_path = os.path.join(os.getcwd(), project_file_name)

    render_args = {
        'project_file_name' : project_file_name,
        'project_class_name' : (opts.name[0].upper() + opts.name[1:]),
        'project_mode' : expand_project_mode_opt(opts.mode),
    }

    for fname, out_fname_template in TEMPLATE_FILES:
        # render template
        template_path = os.path.join(PROJECT_TEMPLATE_PATH, fname)
        with open(template_path) as f:
            template_data = f.read()
        document_data = Template(template_data).render(**render_args)

        # write to output file
        out_fname = out_fname_template.format(project_name=project_file_name)
        out_path = os.path.join(dest_path, out_fname)
        dest_dir = os.path.dirname(out_path)
        try:
            os.makedirs(dest_dir)
        except:
            pass

        with open(out_path, 'w') as f:
            f.write(document_data)
        shutil.copymode(template_path, out_path)

    print('Kaylee project "{}" was created.'.format(
            opts.name))


def _client_template_files(opts):
    template = opts.template

    if template == 'js':
        return ('client/project.js', 'client/{project_name}.js')
    elif template == 'coffee':
        return ('client/project.coffee', 'client/{project_name}.coffee')

def expand_project_mode_opt(opt):
    if opt == 'manual':
        return 'MANUAL_PROJECT_MODE'
    elif opt == 'auto':
        return 'AUTO_PROJECT_MODE'
    else:
        raise ValueError('Invalid project mode option: {}'.format(opt))



from kaylee.testsuite import KayleeTest, load_tests
from kaylee.project import Project, AUTO_PROJECT_MODE, MANUAL_PROJECT_MODE
from kaylee.testsuite.helper import NonAbstractProject

class ProjectTests(KayleeTest):
    def setUp(self):
        pass

    def test_project_mode(self):
        # test if initializing Project without a mode raises a ValueError
        class MyProjectNoMode(NonAbstractProject):
            def __init__(self):
                super(MyProjectNoMode, self).__init__("/script.js")

        self.assertRaises(TypeError, MyProjectNoMode)

        # -- test with invalid mode
        class MyProjectInvalidMode(NonAbstractProject):
            def __init__(self):
                super(MyProjectInvalidMode, self).__init__("/script.js", mode='abc')
        self.assertRaises(ValueError, MyProjectInvalidMode)

        # -- test with valid modes
        class MyProjectWithAutoMode(NonAbstractProject):
            def __init__(self):
                super(MyProjectWithAutoMode, self).__init__("/script.js", AUTO_PROJECT_MODE)
        MyProjectWithAutoMode()

        class MyProjectWithManualMode(NonAbstractProject):
            def __init__(self):
                super(MyProjectWithManualMode, self).__init__("/script.js", MANUAL_PROJECT_MODE)
        MyProjectWithManualMode()



kaylee_suite = load_tests([ProjectTests, ])



# -*- coding: utf-8 -*-
"""
    kaylee.core
    ~~~~~~~~~~~

    This module implements Kaylee's lower level front-end which could
    be easily used with any web framework.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

import sys
import json
import traceback
import logging
from StringIO import StringIO
from functools import partial
from contextlib import closing
from functools import wraps

from .node import Node, NodeID
from .errors import (KayleeError, InvalidResultError, NodeRequestRejectedError,
                     InvalidConfigurationError)
from .controller import KL_RESULT
from .util import DictAsObjectWrapper

log = logging.getLogger(__name__)

#: Returns the results of :function:`json.dumps` in compact encoding
json.dumps = partial(json.dumps, separators=(',', ':'))

ACTION_TASK = 'task'
ACTION_UNSUBSCRIBE = 'unsubscribe'
ACTION_NOP = 'nop'


def json_error_handler(f):
    """A decorator that wraps a function into try..catch block and returns
    JSON-formatted "{ error : str(Exception) }" if an exception has been
    raised.
    """
    #pylint: disable-msg=W0703
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            exc_str = str(e)
            if log.getEffectiveLevel() == logging.DEBUG:
                with closing(StringIO()) as buf:
                    # exc_type, exc_value, exc_traceback = sys.exc_info()
                    exc_traceback = sys.exc_info()[2]
                    traceback.print_tb(exc_traceback,
                                       limit= None,
                                       file= buf)
                    exc_str += '\n' + buf.getvalue()
            return json.dumps({'error': exc_str })

    return wrapper


class Kaylee(object):
    """The Kaylee class serves as a layer between a WSGI server (framework)
    and Kaylee applications. The data flow between Kaylee server and the
    client is kept in JSON format.

    .. note:: It is the job of the WSGI front-end to set the response
              content-type to "application/json".

    See :ref:`loading_kaylee_object` for  Kaylee object initialization and
    loading procedure.

    :param registry: active nodes registry
    :param session_data_manager: global session data manager
    :param applications: a list of applications (:class:`Controller` objects)
    :param \**kwargs: Kaylee configuration arguments.
    :type registry: :class:`NodesRegistry`
    :type session_data_manager: :class:`SessionDataManager` or None
    :type applications: list or None
    """
    def __init__(self, registry, session_data_manager = None,
                 applications = None, **kwargs):
        #: An internal configuration storage object which maintains
        #: the configuration initially parsed from ``**kwargs``.
        #: The options are accessed as object attributes, e.g.:
        #: ``kl.config.SECRET_KEY``.
        self.config = Config(**kwargs)

        #: Active nodes registry (an instance of :class:`NodesRegistry`).
        self.registry = registry

        self.session_data_manager = session_data_manager
        if applications is not None:
            self._applications = Applications(applications)
        else:
            self._applications = Applications.empty()

        log.info(str(self._applications))


    @json_error_handler
    def register(self, remote_host):
        """Registers the remote host (browser) as Kaylee Node and returns
        JSON-formatted data with the following fields:

        * node_id - node id (hex-formatted string)
        * config  - client configuration (see :ref:`settings`).
        * applications - a list of Kaylee applications' names.

        :param remote_host: the IP address of the remote host
        :type remote_host: string
        """
        node = Node(NodeID.for_host(remote_host))
        self.registry.add(node)
        return json.dumps ({ 'node_id' : str(node.id),
                             'config' : self.config.to_client_dict(),
                             'applications' : self._applications.names } )

    @json_error_handler
    def unregister(self, node_id):
        """Removes the node from the nodes registry. Practically this means
        that the remote host (browser) disconnects or is disconnected from
        the Kaylee server.

        :param node_id: a valid node id
        :type node_id: string
        """
        del self.registry[node_id]

    @json_error_handler
    def subscribe(self, node_id, application):
        """Subscribes a node to an application. After a successful subscription
        the node receives a client-side application configuration and invokes
        client-side project initialization routines.

        :param node_id: a valid node id
        :param application: registered Kaylee application name
        :type node_id: string
        :type application: string
        :returns: jsonified node configuration
        """
        try:
            node = self.registry[node_id]
        except KeyError:
            raise KayleeError('Node "{}" is not registered'.format(node_id))

        try:
            app = self._applications[application]
            client_config = node.subscribe(app)
            return json.dumps(client_config)
        except KeyError:
            raise KayleeError('Application "{}" was not found'
                              .format(application))

    @json_error_handler
    def unsubscribe(self, node_id):
        """Unsubscribes the node from the bound application.

        :param node_id: a valid node id.
        :type node_id: string
        """
        self.registry[node_id].unsubscribe()

#    @json_error_handler
    def get_action(self, node_id):
        """Returns an action (usually a task from the subscribed application).
        The format of the JSON response is::

          {
              'action': <action>,
              'data': <data>
          }

        Here, <action> tells the Node, what should it do and <data> is
        the attached data. The valid <actions> are:

        * **"task"** - indicated that <data> contains task data
        * **"unsubscribe"** - indicates that Kaylee server has unsubscribed
          the Node from the application. Any further action request by the
          node raises :class:`NodeNotSubscribedError
          <kaylee.error.NodeNotSubscribedError>`.
        * **"nop"** - indicates that no operation should be carried out by
          the node right now.

        :param node_id: a valid node id
        :type node_id: string
        """
        node = self.registry[node_id]
        try:
            task = node.get_task()
            self._store_session_data(node, task)
            # update node before returning a task
            if node.dirty:
                self.registry.update(node)
                node.dirty = False
            return self._json_action(ACTION_TASK, task)
        except NodeRequestRejectedError as e:
            return self._json_action(ACTION_UNSUBSCRIBE,
                                     'The node has been automatically '
                                     'unsubscribed: {}'.format(e))

    @json_error_handler
    def accept_result(self, node_id, result):
        """Accepts the results from the node. Returns the next action if
        :config:`AUTO_GET_ACTION` configuration option is True. Otherwise
        returns the "nop" (no operatiotion) action.

        :param node_id: a valid node id
        :param result: the result returned by the node.
        :type node_id: string
        :type result: string with JSON-encoded dict data.
        :returns: A task (an action) returned by :meth:`get_action` or
                 "nop" action.
        """
        node = self.registry[node_id]
        try:
            if not isinstance(result, basestring):
                raise ValueError('Kaylee expects the incoming result to be in '
                                 'string format, not {}'.format(
                                     result.__class__.__name__))
            parsed_result = json.loads(result)
            if not isinstance(parsed_result, dict):
                raise ValueError('The returned result was not parsed '
                                 'as dict: {}'.format(parsed_result))
            self._restore_session_data(node, parsed_result)
            node.accept_result(parsed_result)
        except InvalidResultError as e:
            self.unsubscribe(node)
            raise e

        #pylint: disable-msg=E1101
        if self.config.AUTO_GET_ACTION:
            return self.get_action(node.id)
        return self._json_action(ACTION_NOP)

    def clean(self):
        """Removes the outdated nodes from Kaylee's nodes storage."""
        self.registry.clean()

    def _store_session_data(self, node, task):
        if self.session_data_manager is not None:
            self.session_data_manager.store(node, task)

    def _restore_session_data(self, node, result):
        if not KL_RESULT in result:
            if self.session_data_manager is not None:
                self.session_data_manager.restore(node, result)

    @property
    def applications(self):
        """Available applications container (
        :class:`kaylee.core.Applications` object)"""
        return self._applications

    @staticmethod
    def _json_action(action, data = ''):
        return json.dumps( { 'action' : action, 'data' : data } )


class Config(DictAsObjectWrapper):
    """The ``Config`` object maintains the run-time Kaylee
    configuration options (see :ref:`configuration` for full description).
    """
    client_config_fields = [
        'AUTO_GET_ACTION',
    ]

    def __init__(self, **kwargs):
        super(Config, self).__init__(**kwargs)
        self._dirty = True
        self._cached_dict = {}
        self._validate()

    def _validate(self):
        #pylint: disable-msg=E1101
        if not isinstance(self.AUTO_GET_ACTION, bool):
            raise InvalidConfigurationError('AUTO_GET_ACTION is not a boolean')
        #pylint: disable-msg=E1101
        if not isinstance(self.SECRET_KEY, basestring):
            raise InvalidConfigurationError('SECRET_KEY is not a string object')

    def __setattr__(self, name, value):
        if name != '_dirty':
            self.__dict__[name] = value
            self.__dict__['_dirty'] = True
        else:
            self.__dict__[name] = value

    def to_client_dict(self):
        if self._dirty:
            self._cached_dict = { k : getattr(self, k)
                                  for k in self.client_config_fields }
            self._dirty = False
        return self._cached_dict


class Applications(object):
    """A readonly container for active Kaylee applications.

    :param controllers: A list of :class:`Controller` objects.
    """
    def __init__(self, controllers):
        self._controllers = {c.name : c for c in controllers}

        #: A list of apllications' names
        self.names = sorted(self._controllers.keys())

    def __getitem__(self, name):
        """Gets an application (an instance of :class:`Controller`)
        by its name.
        """
        return self._controllers[name]

    def __contains__(self, name):
        """Checks if the container contains application with requested
        name."""
        return name in self._controllers

    def __len__(self):
        """Returns the amount of applications in the container."""
        return len(self._controllers)

    def __str__(self):
        s = 'Kaylee applications({}): '.format(len(self))
        s += '[{}]'.format(', '.join(cname for cname in self._controllers))
        return s



    @staticmethod
    def empty():
        return Applications([])



from django.conf.urls import patterns, url
from kaylee.controller import app_name_pattern
from kaylee.node import node_id_pattern

urlpatterns = patterns('kaylee.contrib.frontends.django_frontend.views',
    url(r'^register$', 'register_node'),
    url(r'^apps/(?P<app_name>{})/subscribe/(?P<node_id>{})$'
        .format(app_name_pattern, node_id_pattern), 'subscribe_node'),
    url(r'^actions/(?P<node_id>{})$'.format(node_id_pattern), 'actions'),
)



# Indicates whether Kaylee will automatically return a next action
# when a result is accepted from a node.
AUTO_GET_ACTION = True

# The key used for session encryption etc.
SECRET_KEY = '{{ SECRET_KEY }}'

# A directory in which Kaylee searches for user projects
PROJECTS_DIR = '{{ PROJECTS_DIR }}'

# Nodes registry configuration
REGISTRY = {
   'name' : 'MemoryNodesRegistry',
   'config' : {
       'timeout' : '30m'
   },
}

# Session data manager configuration
SESSION_DATA_MANAGER = {
    'name' : 'NodeSessionDataManager',
    'config' : {}
}



# Add the applications' configurations here
APPLICATIONS = [

]



#pylint: disable-msg=W0402,W0212,E0202
#W0402: 15,0: Uses of a deprecated module 'string'
#W0212: Access to a protected member _wrapped of a client class
#E0202: KayleeUtilTests.test_lazy_object.NonLazy.x: An attribute
#       affected in kaylee.testsuite.util_tests line 48 hide this method
#       FALSE ALARM
###

import string
import kaylee
from kaylee.testsuite import KayleeTest, load_tests
from kaylee.util import (parse_timedelta, LazyObject, random_string,
                         get_secret_key, DictAsObjectWrapper,
                         RecursiveDictAsObjectWrapper)
from kaylee import KayleeError

class KayleeUtilTests(KayleeTest):
    def setUp(self):
        kaylee.setup(None)

    def test_parse_timedelta(self):
        t1 = parse_timedelta('2d')
        self.assertEqual(t1.days, 2)
        self.assertEqual(t1.seconds, 0)
        self.assertEqual(t1.microseconds, 0)

        t2 = parse_timedelta('2d 1h')
        self.assertEqual(t2.days, 2)
        self.assertEqual(t2.seconds, 3600)

        t3 = parse_timedelta('1h 10m 10s')
        self.assertEqual(t3.days, 0)
        self.assertEqual(t3.seconds, 4210)
        self.assertEqual(t3.microseconds, 0)

        t4 = parse_timedelta('25h 10s')
        self.assertEqual(t4.days, 1)
        self.assertEqual(t4.seconds, 3610)

        t5 = parse_timedelta('2d 1x')
        self.assertEqual(t5.days, 2)
        self.assertEqual(t5.seconds, 0)

        self.assertRaises(KayleeError, parse_timedelta, '25x 10x')

    def test_lazy_object(self):
        #pylint: disable-msg=W0201
        #W0201: lo.y defined outside __init__

        class NonLazy(object):
            def __init__(self):
                self._x = 10
                self.y = 20

            @property
            def x(self):
                return self._x

            @x.setter
            def x(self, val):
                self._x = val

            def x_plus(self, val):
                self.x += val

        class MyLazyObject(LazyObject):
            def _setup(self, obj = None):
                self._wrapped = NonLazy() if obj is None else obj

        lo = MyLazyObject()
        self.assertIsNone(lo._wrapped)
        lo.y = 20
        self.assertIsNotNone(lo._wrapped)
        self.assertEqual(lo.y, 20)
        self.assertEqual(lo.x, 10)
        lo.x_plus(40)
        self.assertEqual(lo.x, 50)

    def test_random_string(self):
        # test length
        s = random_string(5)
        self.assertEqual(len(s), 5)
        s = random_string(0)
        self.assertEqual(len(s), 0)
        s = random_string(100)
        self.assertEqual(len(s), 100)

        # test 'alphabet' argument
        s = random_string(10, alphabet='a')
        self.assertEqual(s, 'a'*10)

        # create sets for less resource-demanding testing
        _string_digits = set(string.digits)
        _string_ascii_uppercase = set(string.ascii_uppercase)
        _string_ascii_lowercase = set(string.ascii_lowercase)
        # test other arguments
        s = random_string(1000, uppercase=False, lowercase=False, special=False)
        for c in s:
            self.assertIn(c, _string_digits)

        s = random_string(1000, lowercase=False, digits=False, special=False)
        for c in s:
            self.assertIn(c, _string_ascii_uppercase)

        s = random_string(1000, uppercase=False, digits=False, special=False)
        for c in s:
            self.assertIn(c, _string_ascii_lowercase)

        # test 'extra' argument
        extra = set('123')
        s = random_string(1000, uppercase=False, lowercase=False, digits=False,
                          special=False, extra=''.join(extra))
        for c in s:
            self.assertIn(c, extra)

        extra = set('1234567890')
        s = random_string(1000, uppercase=False, digits=False, special=False,
                          extra=''.join(extra))
        check_set = _string_ascii_lowercase.union(extra)
        for c in s:
            self.assertIn(c, check_set)

        special = set('!@#$%^&*()_-+=?/><,.|":;`~')
        s = random_string(1000, lowercase=False, uppercase=False, digits=False)
        for c in s:
            self.assertIn(c, special)

    def test_get_secret_key(self):
        sk = get_secret_key('abc')
        self.assertEqual(sk, 'abc')

        # test when config is not loaded
        self.assertRaises(KayleeError, get_secret_key)

        # test loading from config
        from kaylee.testsuite import test_settings
        from kaylee import setup
        setup(test_settings)
        sk = get_secret_key()
        self.assertEqual(sk, test_settings.SECRET_KEY)

        # test if default parameter works after previous call
        sk = get_secret_key('abc')
        self.assertEqual(sk, 'abc')

        # test for proper behaviour after releasing the object from proxy
        setup(None)
        self.assertRaises(KayleeError, get_secret_key)

        # and the final test :)
        sk = get_secret_key('abc')
        self.assertEqual(sk, 'abc')

    def test_dict_as_object_wrapper(self):
        #pylint: disable-msg=E1101
        #E1101 Instance of 'DictAsObjectWrapper' has no 'A' member
        d = {'A' : 10, 'B' : 20}
        wo = DictAsObjectWrapper(**d)
        self.assertEqual(wo.A, 10)
        self.assertEqual(wo.B, 20)

        d = {'A' : 10, 'B' : {'C' : 30}}
        wo = DictAsObjectWrapper(**d)
        self.assertEqual(wo.B, {'C' : 30})

        wo = RecursiveDictAsObjectWrapper(**d)
        self.assertEqual(wo.B.C, 30)

kaylee_suite = load_tests([KayleeUtilTests, ])



# -*- coding: utf-8 -*-
"""
    kaylee.controller
    ~~~~~~~~~~~~~~~~~

    This module provides classes for Kaylee controllers.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

import re
from abc import ABCMeta, abstractmethod

#: The Application name regular expression pattern which can be used in
#: e.g. web frameworks' URL dispatchers.
app_name_pattern = r'[a-zA-Z\.\d_-]+'

#: Indicates active state of an application.
ACTIVE = 0x2
#: Indicates completed state of an application.
COMPLETED = 0x4

KL_RESULT = '__klr__'

#: The ``NO_SOLUTION`` result returned by the node indicates that no solution
#: was found for the given task. The controller must take this information
#: into account.
NO_SOLUTION = { KL_RESULT : 0x2 }

#: The ``NOT_SOLVED`` result returned by the node indicates that the current
#: task was simply not solved for some reason and there are no results to
#: accept.
NOT_SOLVED = { KL_RESULT : 0x4 }


class Controller(object):
    """A Controller object maintains the data (tasks and the results) flow
    between a project the outer :class:`Kaylee` interface and the storages.
    An instance of ``Controller`` with a bound project and storages forms
    a *Kaylee Application*.

    :param name: application name
    :param project: bound project
    :param permanent_storage: permanent application results storage
    :param temporal_storage: internal storage for storing intermediate
                             (temporal) results
    :type name: string
    :type project: :class:`Project`
    :type permanent_storage: :class:`PermanentStorage`
    :type temporal_storage: :class:`TemporalStorage`
    """
    __metaclass__ = ABCMeta

    _app_name_re = re.compile('^{}$'.format(app_name_pattern))

    def __init__(self, name, project, permanent_storage,
                 temporal_storage=None):
        if Controller._app_name_re.match(name) is None:
            raise ValueError('Invalid application name: {}'
                             .format(name))
        self.name = name
        self.project = project
        self.permanent_storage = permanent_storage
        self.temporal_storage = temporal_storage
        self._state = ACTIVE

    @abstractmethod
    def get_task(self, node):
        """Returns a task for the node.

        :param node: Kaylee Node requesting the task for computation.
        :type node: :class:`Node`
        :throws ApplicationCompletedError: if the application is completed.
        """

    @abstractmethod
    def accept_result(self, node, result):
        """Accepts and processes the results from a node.

        :param node: Kaylee Node from which the results have been received.
        :param result: JSON-parsed task results.
        :type result: :class:`dict` or :class:`list`
        """

    def store_result(self, task_id, result):
        """Stores the result to permanent storage and notifies the bound
        project."""
        self.permanent_storage.add(task_id, result)
        self.project.result_stored(task_id, result, self.permanent_storage)

    @property
    def completed(self):
        """Indicates whether the application is completed."""
        return self._state & COMPLETED == COMPLETED

    @completed.setter
    def completed(self, val):
        if val:
            self._state |= COMPLETED
        else:
            self._state &= ~COMPLETED

    def __hash__(self):
        return hash(self.name)



from copy import deepcopy
from kaylee.testsuite import KayleeTest, load_tests
from kaylee.node import Node, NodeID
from kaylee import KayleeError
from kaylee.session import (_encrypt, _decrypt, JSONSessionDataManager,
                            NodeSessionDataManager, PhonySessionDataManager,
                            SESSION_DATA_ATTRIBUTE)

class KayleeSessionTests(KayleeTest):
    def test_encrypt_decrypt(self):
        d1 = {'f1' : 'val1', 'f2' : 20}
        s1 = _encrypt(d1, secret_key='abc')
        d1_d = _decrypt(s1, secret_key='abc')
        self.assertEqual(d1, d1_d)

        # test if incorrect signature raises KayleeError
        s2 = s1[3:] # pad the signature
        self.assertRaises(KayleeError, _decrypt, s2, secret_key='abc')

    def test_node_session_data_manager(self):
        node = Node(NodeID.for_host('127.0.0.1'))
        task = {
            'id' : 'i1',
            '#s1' : 10,
            '#s2' : [1, 2, 3],
        }

        nsdm = NodeSessionDataManager()
        self.assertIsNone(node.session_data)
        nsdm.store(node, task)
        self.assertIsNotNone(node.session_data)

        result = {
            'res' : 'someres',
        }
        nsdm.restore(node, result)
        expected_restored_result = {
            'res' : 'someres',
            '#s1' : 10,
            '#s2' : [1, 2, 3],
        }
        self.assertEqual(result, expected_restored_result)
        self.assertIsNone(node.session_data)


        # test that session is not store in case of no session variables
        task = {
            'id' : 'i2'
        }

        orig_task = deepcopy(task)
        nsdm.store(node, task)
        self.assertIsNone(node.session_data)

        nsdm.restore(node, task)
        self.assertEqual(task, orig_task)


    def test_json_session_data_manager(self):
        node = Node(NodeID.for_host('127.0.0.1'))
        task = {
            'id' : 'i1',
            '#s1' : 10,
            '#s2' : [1, 2, 3],
        }

        jsdm = JSONSessionDataManager(secret_key='abc')
        jsdm.store(node, task)
        self.assertIn(SESSION_DATA_ATTRIBUTE, task)
        self.assertEqual(task['id'], 'i1')
        self.assertNotIn('#s1', task)
        self.assertNotIn('#s2', task)

        result = {
            'res' : 'someres',
            SESSION_DATA_ATTRIBUTE : task[SESSION_DATA_ATTRIBUTE],
        }
        jsdm.restore(node, result)

        expected_restored_result = {
            'res' : 'someres',
            '#s1' : 10,
            '#s2' : [1, 2, 3],
        }
        self.assertEqual(result, expected_restored_result)

        # test that session is not store in case of no session variables
        task = {
            'id' : 'i2'
        }

        orig_task = deepcopy(task)
        jsdm.store(node, task)
        self.assertEqual(task, orig_task)

        jsdm.restore(node, task)
        self.assertEqual(task, orig_task)

    def test_phony_session_data_manager(self):
        node = Node(NodeID.for_host('127.0.0.1'))
        task1 = {
            'id' : 'i1',
            '#s1' : 10,
            '#s2' : [1, 2, 3],
        }

        psdm = PhonySessionDataManager()
        self.assertRaises(KayleeError, psdm.store, node, task1)

        task2 = {
            'id' : 'i1',
        }
        # this should be executed without any errors
        self.assertIsNone(psdm.store(node, task2))


kaylee_suite = load_tests([KayleeSessionTests, ])



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod

from kaylee.testsuite import (load_tests, TestPermanentStorage)
from kaylee.testsuite.helper import SubclassTestsBase
from kaylee.testsuite.projects.auto_test_project import AutoTestProject
from kaylee.node import Node, NodeID
from kaylee.contrib.controllers import SimpleController
from kaylee.errors import InvalidResultError



class ControllerTestsBase(SubclassTestsBase):
    """The class contains only basic generic tests that can be applied
    to any controller. Please subclass and extend the tests for thorough
    testing.
    """
    __metaclass__ = ABCMeta

    def setUp(self):
        super(ControllerTestsBase, self).setUp()

    @abstractmethod
    def test_init(self):
        pass

    def test_get_task(self):
        node, ctr = self.make_node_and_controller()
        task = ctr.get_task(node)
        self.assertIsInstance(task, dict)
        self.assertIn('id', task)
        self.assertTrue(len(str(task['id']).strip()) > 0)

        #pylint: disable-msg=W0612
        for i in range(0, self.MANY):
            task = ctr.get_task(node)
            self.assertTrue(task is None or isinstance(task, dict))

    def test_accept_result(self):
        node, ctr = self.make_node_and_controller()
        task = ctr.get_task(node)
        res = { 'res' : task['id'] }
        ctr.accept_result(node, res)
        self.assertRaises(InvalidResultError, ctr.accept_result, node, {})

    def make_node_and_controller(self):
        ctr = self.cls_instance()
        n = Node(NodeID())
        n.subscribe(ctr)
        return n, ctr


class SimpleControllerTests(ControllerTestsBase):
    def test_init(self):
        pass

    def cls_instance(self):
        return SimpleController('test_simple_controller_app',
                                AutoTestProject(),
                                TestPermanentStorage())


kaylee_suite = load_tests([SimpleControllerTests,])



from .start_env import StartEnvCommand
from .run import RunCommand
from .start_project import StartProjectCommand
from .build import BuildCommand

commands_classes = [
    StartEnvCommand,
    StartProjectCommand,
    RunCommand,
    BuildCommand,
]



from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from django.http import HttpResponse

from kaylee import kl

def register_node(request):
    reg_data = kl.register(request.META['REMOTE_ADDR'])
    return json_response(reg_data)

#pylint: disable-msg=W0613
#W0613:  Unused argument 'request'
@csrf_exempt
@require_http_methods(["POST"])
def subscribe_node(request, app_name, node_id):
    node_config = kl.subscribe(node_id, app_name)
    return json_response(node_config)

@csrf_exempt
def actions(request, node_id):
    if request.method == 'GET':
        return json_response( kl.get_action(node_id) )
    elif request.method == 'POST':
        next_task = kl.accept_result(node_id, request.raw_post_data)
        return json_response(next_task)

def json_response(s):
    return HttpResponse(s, content_type = 'application/json')



from __future__ import print_function
import os
from argparse import ArgumentTypeError
from kaylee.server import run
from kaylee.manager import LocalCommand


def port_type(port):
    try:
        p = int(port)
        if not 1024 < p <= 65536:
            msg = 'The available port numbers are in (1024..65536] range.'
            raise ArgumentTypeError(msg)
        return p
    except ValueError:
        raise '{} is not a valid port'.format(port)


class RunCommand(LocalCommand):
    name = 'run'
    help = 'Runs Kaylee development/testing server'

    args = {
        ('-s', '--settings-file') : dict(default='settings.py'),
        ('-b', '--build-dir') : dict(default='_build'),
        ('-p', '--port') : dict(default='5000',
                                type=port_type,
                                help='Web server port number'),
        '--debug' : dict(default=False,
                         action='store_true',
                         help='Debug ON'),
    }

    @staticmethod
    def execute(opts):
        validate_settings_file(opts)
        validate_build_dir(opts)
        run_dev_server(opts)


def validate_settings_file(opts):
    if not os.path.exists(opts.settings_file):
        raise OSError('Cannot find the settings file "{}"'
                      .format(opts.settings_file))


def validate_build_dir(opts):
    if not os.path.exists(opts.build_dir):
        raise OSError (
            'Cannot find build directory "{}". \n'
            'Have you forgotten building the environment? \n'
            'If not, please specify with -b or --build-dir.'
            .format(opts.build_dir))


def run_dev_server(opts):
    print('Launching Kaylee development/testing server...')
    run(opts.settings_file, opts.build_dir, opts.port, opts.debug)



# -*- coding: utf-8 -*-

from kaylee.testsuite import  PROJECTS_DIR

AUTO_GET_ACTION = True

PROJECTS_DIR = PROJECTS_DIR

SECRET_KEY = 'aJD2fn;1340913)*(!!&$)(#&<AHFB12b'

REGISTRY = {
    'name' : 'MemoryNodesRegistry',
    'config' : {
        'timeout' : '2s',
        },
}

APPLICATIONS = [
    { 'name' : 'test.1',
      'description' : 'Test application',
      'project' : {
            'name' : 'AutoTestProject',
            },
      'controller' : {
            'name' :'TestController1',
            'config' : {},
            'temporal_storage' : {
                'name' : 'MemoryTemporalStorage',
                },
            'permanent_storage' : {
                'name' : 'MemoryPermanentStorage',
                },
            },
      }
    ]



# -*- coding: utf-8 -*-
"""
    kaylee.manager.command
    ~~~~~~~~~~~~~~~~~~~~~~

    The module contains base Kaylee manager commands.

    :copyright: (c) 2013 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""
import argparse


class BaseCommand(object):
    #: Command help text
    name = ''
    help = ''
    args = {}

    @classmethod
    def add_sub_parser(cls, sub_parsers_object):
        if cls.name.strip() == '':
            raise ValueError('{}.name is empty.'.format(cls.__name__))

        parser = sub_parsers_object.add_parser(cls.name, help=cls.help)
        for arg, arg_opts in cls.args.items():
            if isinstance(arg, str):
                arg = [arg, ]
            parser.add_argument(*arg, **arg_opts)
        parser.set_defaults(handler=cls.execute)

    @staticmethod
    def execute(ns):
        raise NotImplementedError('The command has no execute() static method.')


class AdminCommand(BaseCommand):
    #pylint: disable-msg=W0223
    #W0223: Method 'execute' is abstract in class 'BaseCommand'
    #       but is not overridden (throws NotImplementedError).
    pass


class LocalCommand(BaseCommand):
    #pylint: disable-msg=W0223
    #W0223: Method 'execute' is abstract in class 'BaseCommand'
    #       but is not overridden (throws NotImplementedError).
    pass


class CommandsManager(object):
    help = ''
    command_class = BaseCommand

    def __init__(self):
        self.parser = argparse.ArgumentParser(description=self.help)
        self.sub_parsers = self.parser.add_subparsers(
            help='Commands.')
        # add sub-commands
        from .commands import commands_classes
        for cmd_cls in commands_classes:
            if issubclass(cmd_cls, self.command_class):
                self.add_command(cmd_cls)

    def add_command(self, cmd_cls):
        cmd_cls.add_sub_parser(self.sub_parsers)

    def parse(self, argv=None):
        ns = self.parser.parse_args(argv)
        if 'handler' in ns:
            ns.handler(ns)

    @classmethod
    def execute_from_command_line(cls):
        try:
            cls().parse()
        except Exception as e:
            print(e)
            raise SystemExit(1)


class AdminCommandsManager(CommandsManager):
    help = 'Kaylee admin'
    command_class = AdminCommand


class LocalCommandsManager(CommandsManager):
    help = 'Kaylee local environment manager'
    command_class = LocalCommand



# -*- coding: utf-8 -*-

from flask import Blueprint, request, Response
from kaylee import kl

bp = Blueprint('kaylee_blueprint', __name__)
kaylee_blueprint = bp # just an alias for importing convenience

@bp.route('/register')
def register_node():
    reg_data = kl.register(request.remote_addr)
    return json_response(reg_data)

@bp.route('/apps/<app_name>/subscribe/<node_id>', methods=['POST'])
def subscribe_node(node_id, app_name):
    node_config = kl.subscribe(node_id, app_name)
    return json_response(node_config)

@bp.route('/actions/<node_id>', methods=['GET', 'POST'])
def tasks(node_id):
    if request.method == 'GET':
        return json_response(kl.get_action(node_id))
    else:
        next_task = kl.accept_result(node_id, request.data)
        # the reason for using request.data instead of request.json
        # is that Kaylee expects the "raw", non-processed data
        return json_response(next_task)

def json_response(s):
    return Response(s, mimetype = 'application/json')



# -*- coding: utf-8 -*-
"""
    kaylee.contrib.frontends
    ~~~~~~~~~~~~~~~~~~~~~~~~

    The sub-package contains various frond-ends integration code.

    :copyright: (c) 2013 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""



# -*- coding: utf-8 -*-
"""
    kaylee.util
    ~~~~~~~~~~~

    The module contains common Kaylee routines and classes.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""
#pylint: disable-msg=W0402,W0212,R0913
#W0402: 15,0: Uses of a deprecated module 'string'
#W0212: Access to a protected member _wrapped of a client class
#R0913: random_string: Too many arguments (7/5)
###

import os
import re
import sys
import random
import string
import importlib
import contextlib
import logging
from datetime import timedelta
from .errors import KayleeError


def parse_timedelta(s):
    try:
        match = parse_timedelta.timeout_regex.match(s)
    except AttributeError:
        parse_timedelta.timeout_regex = re.compile(
            r'((?P<days>\d+?)d)?\s?((?P<hours>\d+?)h)?\s?'
            r'((?P<minutes>\d+?)m)?\s?((?P<seconds>\d+?)s)?')
        match = parse_timedelta.timeout_regex.match(s)

    try:
        time_params = {}
        for (name, param) in match.groupdict().iteritems():
            if param is not None:
                time_params[name] = int(param)
        if time_params == {}:
            raise Exception()
        return timedelta(**time_params)
    except:
        raise KayleeError('Wrong timedelta string: {}'.format(s))


def import_object(name):
    modname, objname = name.rsplit('.', 1)
    mod = importlib.import_module(modname)
    try:
        return getattr(mod, objname)
    except AttributeError:
        raise ImportError('Object {} was not found in module {}'
                          .format(objname, modname))


def _new_method_proxy(func):
    def inner(self, *args):
        try:
            return func(self._wrapped, *args)
        except AttributeError:
            self._setup()
            return func(self._wrapped, *args)
    return inner


class LazyObject(object):
    """
    A wrapper for another class that can be used to delay instantiation of the
    wrapped class.

    By subclassing, you have the opportunity to intercept and alter the
    instantiation.
    """
    def __init__(self):
        self._wrapped = None

    __getattr__ = _new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == '_wrapped':
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__['_wrapped'] = value
        else:
            try:
                setattr(self._wrapped, name, value)
            except AttributeError:
                self._setup()
                setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == '_wrapped':
            raise TypeError('Cannot delete _wrapped.')
        try:
            delattr(self._wrapped, name)
        except AttributeError:
            self._setup()
            delattr(self._wrapped, name)

    def _setup(self, obj = None):
        """
        Must be implemented by sub-classes to initialize the wrapped object.

        :param obj: An optional object to wrap. Note that checking the object
                    type and wrapping it must be done in the sub class as well.
        """
        raise NotImplementedError

    # introspection support:
    __dir__ = _new_method_proxy(dir)


class DictAsObjectWrapper(object):
    def __init__(self, **kwargs):
        for key, val in kwargs.iteritems():
            setattr(self, key, val)


class RecursiveDictAsObjectWrapper(object):
    def __init__(self, **kwargs):
        def _update(obj, d):
            for key, val in d.iteritems():
                if isinstance(val, dict):
                    nested_obj = RecursiveDictAsObjectWrapper(**val)
                    setattr(obj, key, nested_obj)
                else:
                    setattr(obj, key, val)
        _update(self, kwargs)


def random_string(length, alphabet=None, lowercase=True, uppercase=True,
                  digits=True, special=True, extra=''):
    if alphabet is None:
        src = extra
        if lowercase:
            src += string.ascii_lowercase
        if uppercase:
            src += string.ascii_uppercase
        if digits:
            src += string.digits
        if special:
            src += '!@#$%^&*()_-+=?/><,.|":;`~'
    else:
        src = alphabet
    return ''.join(random.choice(src) for x in xrange(length))


def get_secret_key(key=None):
    if key is not None:
        return key
    else:
        from kaylee import kl
        if kl._wrapped is None:
            raise KayleeError('Cannot locate a valid secret key because '
                              'Kaylee proxy object has not beet set up.')
        key = kl.config.SECRET_KEY
        if key is None:
            raise KayleeError('SECRET_KEY configuration option is not '
                              'defined.')
        return key



def is_strong_subclass(C, B):
    """Return whether class C is a subclass (i.e., a derived class) of class B
    and C is not B.
    """
    return issubclass (C, B) and C is not B


@contextlib.contextmanager
def nostdout(stdout=True, stderr=True):
    savestderr = sys.stderr
    savestdout = sys.stdout
    class Devnull(object):
        def write(self, _):
            pass
    if stdout:
        sys.stdout = Devnull()
    if stderr:
        sys.stderr = Devnull()
    yield
    sys.stdout = savestdout
    sys.stderr = savestderr


def ensure_dir(path):
    try:
        os.makedirs(path)
    except OSError:
        if not os.path.isdir(path):
            raise

def setup_logging(loglevel=logging.INFO):
    log_format = '%(levelname)s[%(name)s]: %(message)s'
    logging.basicConfig(level=loglevel, format=log_format)



#!/usr/bin/env python

import os
from setuptools import setup

import kaylee


def fullsplit(path, result=None):
    """Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


EXCLUDE_FROM_PACKAGES = [
    'kaylee.manager.commands.templates.project_template',
    'kaylee.testsuite.command_manager_tests_resources.pi_calc',
    'kaylee.testsuite.command_manager_tests_resources.monte_carlo_pi',
]


def is_package(package_name):
    for pkg in EXCLUDE_FROM_PACKAGES:
        if package_name.startswith(pkg):
            return False
    return True

root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)
kaylee_dir = 'kaylee'

# Compile the list of packages available, because distutils doesn't have
# an easy way to do this.
packages, package_data = [], {}

for dirpath, dirnames, filenames in os.walk(kaylee_dir):
    # Ignore PEP 3147 cache dirs and those whose names start with '.'
    dirnames[:] = [d for d in dirnames if not d.startswith('.') and d != '__pycache__']
    parts = fullsplit(dirpath)
    package_name = '.'.join(parts)
    if '__init__.py' in filenames and is_package(package_name):
        packages.append(package_name)
    elif filenames:
        relative_path = []
        while '.'.join(parts) not in packages:
            relative_path.append(parts.pop())
        relative_path.reverse()
        path = os.path.join(*relative_path)
        package_files = package_data.setdefault('.'.join(parts), [])
        package_files.extend([os.path.join(path, f) for f in filenames])

setup(
    name='Kaylee',
    version=kaylee.__version__,
    url='http://github.com/basicwolf/kaylee',
    license='LICENSE',
    author='Zaur Nasibov',
    author_email='zaur@znasibov.info',
    description='A distributed and crowd computing framework',
    long_description=open('README').read(),
    packages=packages,
    package_data=package_data,
    zip_safe=False,
    scripts=[
        'kaylee/bin/kaylee-admin.py',
    ],
    install_requires=[
        'Werkzeug>=0.6.1',
        'Jinja2>=2.4',
        'pycrypto>=2.6',
    ],

    test_suite='kaylee.testsuite.suite',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: JavaScript',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Distributed Computing',
    ],
)



from __future__ import print_function
import os
import sys
import imp
import importlib
import subprocess
import shutil
import kaylee
from kaylee.manager import LocalCommand
from kaylee.loader import find_packages, get_classes_from_module
from kaylee.util import ensure_dir


class BuildCommand(LocalCommand):
    name = 'build'
    help = 'Builds Kaylee environment'

    args = {
        ('-s', '--settings-file') : dict(default='settings.py'),
        ('-b', '--build-dir') : dict(default='_build')
    }

    @staticmethod
    def execute(opts):
        verify_settings(opts)
        verify_build_dir(opts)

        print('Building Kaylee environment...')
        settings = imp.load_source('settings', opts.settings_file)
        build_kaylee(opts)
        build_projects(settings, opts)


def verify_settings(opts):
    if not os.path.exists(opts.settings_file):
        raise OSError('Cannot find the settings file "{}"'
                      .format(opts.settings_file))

def verify_build_dir(opts):
    try:
        if not os.path.exists(opts.build_dir):
            os.mkdir(opts.build_dir)
    except OSError as e:
        raise OSError('Failed to create build directory {}: {}'
                      .format(opts.build_dir, e))


def build_kaylee(opts):
    print('* Copying Kaylee test server files...')
    KLCL_TEMPLATE_DIR = 'templates/build_template'
    KLCL_TEMPLATE_PATH = os.path.join(os.path.dirname(__file__),
                                      KLCL_TEMPLATE_DIR)

    KLCL_CLIENT_PATH = os.path.join(os.path.dirname(kaylee.__file__),
                                    'client')
    CLIENT_FILES = [
        ('kaylee.js', 'kaylee/js/kaylee.js'),
        ('klworker.js', 'kaylee/js/klworker.js'),
    ]

    TEMPLATE_FILES = [
        ('css/klconsole.css', 'kaylee/css/klconsole.css'),
        ('js/kldebug.js', 'kaylee/js/kldebug.js'),
        ('js/jquery.min.js', 'kaylee/js/jquery.min.js'),
        ('index.html', 'index.html'),
    ]

    dest_path = os.path.join(os.getcwd(), opts.build_dir)
    ensure_dir(os.path.join(dest_path, 'kaylee/js'))
    ensure_dir(os.path.join(dest_path, 'kaylee/css'))

    for fname, out_fname in CLIENT_FILES:
        fpath = os.path.join(KLCL_CLIENT_PATH, fname)
        dest_fpath = os.path.join(dest_path, out_fname)
        shutil.copy(fpath, dest_fpath)

    for fname, out_fname in TEMPLATE_FILES:
        fpath = os.path.join(KLCL_TEMPLATE_PATH, fname)
        dest_fpath = os.path.join(dest_path, out_fname)
        shutil.copy(fpath, dest_fpath)


def build_projects(settings, opts):
    filetype_handlers = [
        coffee_handler,
        script_handler,
        stylesheet_handler,
        data_handler,
    ]

    # Make sure that the command is able to import the projects'
    # Python packages.
    sys.path.insert(0, settings.PROJECTS_DIR)

    for pkg_dir in find_packages(settings.PROJECTS_DIR):
        if not is_kaylee_project_directory(pkg_dir):
            continue
        print('* Building {}...'.format(pkg_dir))
        dest_dir = os.path.join(opts.build_dir, pkg_dir)
        opts.dest_dir = dest_dir
        # clear destination directory before building
        if os.path.isdir(dest_dir):
            shutil.rmtree(dest_dir)
        # build project client by applying appropriate
        # file handlers
        client_dir = os.path.join(pkg_dir, 'client')
        for fname in os.listdir(client_dir):
            fpath = os.path.join(client_dir, fname)
            if not os.path.isfile(fpath):
                continue
            for handler in filetype_handlers:
                if handler(fpath, opts) == True:
                    break

def is_kaylee_project_directory(path):
    # check if 'client' directory exists
    client_dir = os.path.join(path, 'client')
    if not os.path.isdir(client_dir):
        return False
    # check for kaylee.Project subclass
    package_name = path.rsplit('/')[-1]
    pymod = importlib.import_module(package_name)
    for cls in get_classes_from_module(pymod):
        if issubclass(cls, kaylee.project.Project):
            return True
    return False

def coffee_handler(fpath, opts):
    if not fpath.endswith('.coffee'):
        return False

    args = ['coffee', '--bare', '-c', fpath]
    proc = subprocess.Popen(args,
                            close_fds=True,
                            stdin=subprocess.PIPE,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE)
    #pylint: disable-msg=W0612
    #W0612: Unused variable 'stdoutdata'
    stdoutdata, stderrdata = proc.communicate()
    if stderrdata is not '':
        raise Exception('Error compiling .coffee script:\n{}'
                        .format(stderrdata))

    dest_dir = os.path.join(opts.dest_dir, 'js')
    ensure_dir(dest_dir)
    outpath = fpath.rsplit('.coffee', 1)[0] + '.js'
    shutil.move(outpath, dest_dir)
    return True


def script_handler(fpath, opts):
    if not fpath.endswith('.js'):
        return False

    dest_dir = os.path.join(opts.dest_dir, 'js')
    ensure_dir(dest_dir)
    shutil.copy(fpath, dest_dir)
    return True


def stylesheet_handler(fpath, opts):
    if not fpath.endswith('.css'):
        return False

    dest_dir = os.path.join(opts.dest_dir, 'css')
    ensure_dir(dest_dir)
    shutil.copy(fpath, dest_dir)
    return True


def data_handler(fpath, opts):
    dest_dir = os.path.join(opts.dest_dir, 'data')
    ensure_dir(dest_dir)
    shutil.copy(fpath, dest_dir)
    return True



from kaylee.testsuite import KayleeTest, load_tests
from datetime import datetime, timedelta
from kaylee import Node, NodeID
from kaylee.node import extract_node_id
from kaylee import InvalidNodeIDError
from kaylee.testsuite import TestController


class NodeIDTests(KayleeTest):
    def setUp(self):
        pass

    def test_init(self):
        n1 = NodeID()
        # make sure that NodeID is slot-based object
        self.assertFalse(hasattr(n1, '__dict__'))
        self.assertEqual(len(n1.binary), 10)
        self.assertEqual(len(str(n1)), 20)
        n2 = NodeID(node_id = str(n1))
        self.assertEqual(n1, n2)
        n3 = NodeID(node_id = n1)
        self.assertEqual(n1, n3)

    def test_for_host(self):
        n1 = NodeID.for_host('127.0.0.1')
        self.assertFalse(hasattr(n1, '__dict__'))
        self.assertEqual(len(n1.binary), 10)
        self.assertEqual(len(str(n1)), 20)
        self.assertRaises(TypeError, NodeID.for_host)
        self.assertRaises(TypeError, NodeID.for_host, 1000)
        self.assertRaises(TypeError, NodeID.for_host, 100.1)

    def test_extract_node_id(self):
        n  = NodeID.for_host('127.0.0.1')
        n1 = extract_node_id(n)
        self.assertIs(n, n1)
        n2 = extract_node_id(str(n))
        self.assertEqual(n, n2)
        node = Node(n)
        n3 = extract_node_id(node)
        self.assertEqual(n, n3)
        self.assertRaises(InvalidNodeIDError, extract_node_id, 'abc')
        self.assertRaises(TypeError, extract_node_id, 10)
        node = Node(n)
        node.id = '10'
        self.assertRaises(InvalidNodeIDError, extract_node_id, node)

    def test_internal_counter(self):
        #pylint: disable-msg=W0612,W0212
        #W0612: Unused variable 'i'
        #W0212: Access to a protected member _inc of a client class

        ###
        NodeID._inc = 0 # modified for test purposes only
        remote = '127.0.0.1'
        for i in xrange(0, 2**16 - 1):
            n = NodeID.for_host(remote)
        bid = bytearray(n.binary)
        self.assertEqual(bid[4], 0xFF)
        self.assertEqual(bid[5], 0xFE)

        n = NodeID.for_host(remote)
        bid = bytearray(n.binary)
        self.assertEqual(bid[4], 0x0)
        self.assertEqual(bid[5], 0x0)

    def test_compare(self):
        self.assertLess(NodeID(), NodeID())

    def test_parse(self):
        n1 = NodeID.for_host('127.0.0.1')
        n2 = NodeID(str(n1))
        self.assertEqual(n1, n2)
        n2 = NodeID(unicode(n1))
        self.assertEqual(n1, n2)

        n2 = NodeID(n1)
        self.assertEqual(n1, n2)
        self.assertRaises(TypeError, NodeID, 123)
        self.assertRaises(InvalidNodeIDError, NodeID, 'abc')
        self.assertRaises(InvalidNodeIDError, NodeID, u'1'*10)

    def test_dates(self):
        n1 = NodeID.for_host('127.0.0.1')
        t1 = n1.timestamp
        now = datetime.now()
        self.assertTrue(timedelta(seconds = 0) <= now - t1
                        <= timedelta(seconds = 3))

    def test_hashability(self):
        n1 = NodeID.for_host('127.0.0.1')
        n2 = NodeID(n1)
        d = { n1 : 'abc' }
        self.assertIn(n1, d)
        self.assertIn(n2, d)


class NodeTests(KayleeTest):
    def setUp(self):
        pass

    def test_init(self):
        nid = NodeID.for_host('127.0.0.1')
        node = Node(nid)
        self.assertEqual(nid, node.id)
        self.assertRaises(TypeError, Node, 100)
        self.assertRaises(TypeError, Node, 'abc')

    def test_dirty(self):
        nid = NodeID.for_host('127.0.0.1')
        node = Node(nid)
        self.assertFalse(node.dirty)

        node.subscribe(TestController.new_test_instance())
        self.assertTrue(node.dirty)
        node.dirty = False

        node.task_id = 123
        self.assertTrue(node.dirty)
        node.dirty = False

        node.session_data = '123'
        self.assertTrue(node.dirty)
        node.dirty = False

    def test_get_set_properties(self):
        nid = NodeID.for_host('127.0.0.1')
        node = Node(nid)

        node.session_data = 'sd1'
        self.assertEqual(node.session_data, 'sd1')

        ctrl = TestController.new_test_instance()
        node.subscribe(ctrl)
        now = datetime.now()
        self.assertTrue(timedelta(seconds = 0)
                        <= now - node.subscription_timestamp
                        <= timedelta(seconds = 3))
        self.assertEqual(node.controller, ctrl)

        node.task_id = 'tid789'
        now = datetime.now()
        self.assertTrue(timedelta(seconds = 0) <= now - node.task_timestamp
                        <= timedelta(seconds = 3))
        self.assertEqual(node.task_id, 'tid789')


kaylee_suite = load_tests([NodeTests, NodeIDTests, ])



# -*- coding: utf-8 -*-
"""
    kaylee.errors
    ~~~~~~~~~~~~~

    This module implements all exceptions raised by the Kaylee package.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""

import warnings


class KayleeError(Exception):
    """Base class for all Kaylee exceptions."""


class KayleeProjectAssertError(KayleeError):
    """Base class for all Kaylee project-related runtime assertion
    (regular, **not** Python ``asert`` checks) errors."""
    pass


class InvalidNodeIDError(KayleeError):
    """Raised when parsing a node id (e.g. hex string) fails."""
    def __init__(self, node_id):
        KayleeError.__init__(self, '{} is not a valid node id'.format(node_id))


class NodeNotSubscribedError(KayleeError):
    """Raised when a Node requests an action or submits results without
    being subscribed to an application."""
    def __init__(self, node):
        KayleeError.__init__(self, 'Node {} is not subscribed'.format(node))


class InvalidResultError(ValueError, KayleeError):
    """Raised when a result received from the client is not valid (e.g.
    :meth:`Project.normalize_result` fails)."""
    def __init__(self, result, why = ''):
        super(InvalidResultError, self).__init__(
            'Invalid result "{}": {}'.format(result, why))


class NoneResultAssertError(KayleeProjectAssertError, ValueError):
    """Raised by controller if a normalized result returned by
    :meth:`Project.normalize_result` is ``None``."""
    def __init__(self, result):
        super(NoneResultAssertError, self).__init__(
            'A result {} has been normalized to None'.format(result))


class NodeRequestRejectedError(KayleeError):
    """Raised when a controller rejects a task request or result response
    from a node for some particular reason (e.g. if the node has submitted
    the same task result twice without "being asked" to do so)."""
    def __init__(self, message):
        KayleeError.__init__(self, ('The node has been rejected: {}'
                                    .format(message)) )


class ApplicationCompletedError(NodeRequestRejectedError):
    """Raised when a node request is rejected due to the
    :attr:`completed <Project.completed>` state of the application.

    Base class: :class:`NodeRequestRejectedError`."""
    def __init__(self, application):
        self.application = application
        super(ApplicationCompletedError, self).__init__(
            'The application "{}" has been completed.'
            .format(application.name) )


class InvalidConfigurationError(KayleeError):
    """Raised when :class:`Kaylee object <Kaylee>` configuration is
    configured improperly."""
    def __init__(self, message):
        super(InvalidConfigurationError, self).__init__(
            'Invalid configuration: ' + message)


class KayleeWarning(UserWarning):
    pass


def warn(message):
    warnings.warn(message, KayleeWarning, 3)



# -*- coding: utf-8 -*-
"""
    kaylee.contrib.storages
    ~~~~~~~~~~~~~~~~~~~~~~~

    The module provides basic Kaylee storages' implementations.

    :copyright: (c) 2013 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""
#pylint: disable-msg=W0231

from kaylee.storage import TemporalStorage, PermanentStorage
from kaylee.node import NodeID

class MemoryTemporalStorage(TemporalStorage):
    """A simple Python dict-based temporal results storage."""

    def __init__(self):
        self._d = {}
        self.clear()
        self._total_count = 0

    def add(self, task_id, node_id, result):
        d = self._d.get(task_id, {})
        d[node_id.binary] = result
        self._d[task_id] = d
        self._total_count += 1

    def remove(self, task_id, node_id=None):
        if node_id is None:
            deleted_results = self._d.pop(task_id)
            self._total_count -= len(deleted_results)
        else:
            del self._d[task_id][node_id]
            self._total_count -= 1

    def clear(self):
        self._d = {}
        self._total_count = 0

    def __getitem__(self, task_id):
        nr_dict = self._d[task_id]
        return ((NodeID(n), r) for n, r in nr_dict.iteritems())

    def contains(self, task_id, node_id=None, result=None):
        try:
            if result is None and node_id is None:
                return task_id in self._d
            elif result is None:
                return node_id in self._d[task_id]
            else:
                return result in self._d[task_id][node_id]
        except KeyError:
            return False

    @property
    def count(self):
        return len(self._d)

    @property
    def total_count(self):
        return self._total_count

    def keys(self):
        return self._d.iterkeys()

    def values(self):
        def node_result_tuple_generator():
            for nr_dict in self._d.itervalues():
                for nid_res_tuple in nr_dict.iteritems():
                    # in Py3 yield from would be proper here
                    yield nid_res_tuple
        return node_result_tuple_generator()


class MemoryPermanentStorage(PermanentStorage):
    """A simple Python dict-based permanent results storage."""

    def __init__(self):
        self._d = {}
        self._total_count = 0

    def add(self, task_id, result):
        try:
            self._d[task_id].append(result)
        except KeyError:
            self._d[task_id] = [result, ]
        self._total_count += 1

    def __getitem__(self, task_id):
        return self._d[task_id]

    def contains(self, task_id, result = None):
        if result is None:
            return task_id in self._d
        else:
            return task_id in self._d and result in self._d[task_id]

    def keys(self):
        return self._d.iterkeys()

    def values(self):
        return self._d.itervalues()

    @property
    def count(self):
        return len(self._d)

    @property
    def total_count(self):
        return self._total_count



# -*- coding: utf-8 -*-
"""
    kaylee.contrib.controllers
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    The module provides basic Kaylee controller implementations.

    :copyright: (c) 2013 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""
from kaylee.controller import Controller, NO_SOLUTION, NOT_SOLVED
from kaylee.errors import (ApplicationCompletedError,
                           NodeRequestRejectedError,
                           NoneResultAssertError,)


class SimpleController(Controller):
    """
    This is a very basic controller which serves the tasks by
    node requests and passes the accepted results directly to the project.
    Its ``completed`` indicator is set to ``True`` the moment the bound
    project is completed. The controller doesn't use a temporal storage.
    """
    def __init__(self, *args, **kwargs):
        super(SimpleController, self).__init__(*args, **kwargs)
        self._tasks_pool = set()
        self._project_depleted = False

    def get_task(self, node):
        task = self.project.next_task()
        if task is None:
            try:
                tp_id = self._tasks_pool.pop()
                task = self.project[tp_id]
            except KeyError:
                # project depleted and nothing in the pool,
                # looks like the application is completed.
                self.completed = True
                raise ApplicationCompletedError(self)

        task_id = task['id']
        node.task_id = task_id
        self._tasks_pool.add(task_id)
        return task

    def accept_result(self, node, result):
        if result == NO_SOLUTION:
            self._tasks_pool.remove(node.task_id)
            return
        elif result == NOT_SOLVED:
            return

        norm_result = self.project.normalize_result(node.task_id, result)
        if norm_result is None:
            raise NoneResultAssertError(result)

        self.store_result(node.task_id, norm_result)
        self._tasks_pool.remove(node.task_id)
        if self.project.completed:
            self.completed = True



class ResultsComparatorController(Controller):
    """
    This controller is a simple implementation of the "trust no one" idea.
    The intermediate results are collected in a temporal storage until
    their number reaches a user-defined limit. Then the results are pairly
    compared and only if they all match among themselves, a single result
    is stored inside the permanent storage. The results are discarded if they
    don't match and the task is pushed back to the "unsolved tasks" pool.

    :param results_count_threshold: The amount of task results to be collected
                                    before running the comparison routine.
    """
    def __init__(self, *args, **kwargs):
        self._results_count_threshold = kwargs.pop('results_count_threshold')
        super(ResultsComparatorController, self).__init__(*args, **kwargs)
        self._tasks_pool = set()

    def get_task(self, node):
        task = self.project.next_task()
        if task is None:
            try:
                tp_id = self._tasks_pool.pop()
                task = self.project[tp_id]
            except KeyError:
                # project depleted and nothing in the pool,
                # looks like the application has completed.
                self.completed = True
                raise ApplicationCompletedError(self)

        task_id = task['id']
        node.task_id = task_id
        self._tasks_pool.add(task_id)

        if self.temporal_storage.contains(task_id, node.id):
            raise NodeRequestRejectedError('The result of this task has been '
                                           'already accepted.')
        return task

    def accept_result(self, node, result):
        if result == NOT_SOLVED:
            return

        task_id = node.task_id

        if result == NO_SOLUTION:
            norm_result = result
        else:
            norm_result = self.project.normalize_result(task_id, result)
            if norm_result is None:
                raise NoneResultAssertError(result)

        # 'results' computed for the task by various nodes
        tmp_results = self.temporal_storage[task_id]
        if len(tmp_results) == self._results_count_threshold - 1:
            if self._results_are_equal(norm_result, tmp_results):
                del self.temporal_storage[task_id]
                self._tasks_pool.remove(task_id)
                if result != NO_SOLUTION:
                    self.store_result(task_id, norm_result)
            else:
                # Something is wrong with either current result or any result
                # which was received previously. At this point we discard all
                # results associated with task_id and task_id remains in
                # tasks_pool (if the result is not NO_SOLUTION)
                del self.temporal_storage[task_id]
                if result == NO_SOLUTION:
                    self._tasks_pool.remove(task_id)
            node.task_id = None
        else:
            self.temporal_storage.add(task_id, node.id, norm_result)

    @staticmethod
    def _results_are_equal(r0, res):
        #pylint: disable-msg=W0612
        #W0612: Unused variable 'node_id'
        ###
        for node_id, res in res.iteritems():
            if r0 != res:
                return False
        return True

    def store_result(self, task_id, result):
        super(ResultsComparatorController, self).store_result(task_id, result)
        if self.project.completed:
            self.completed = True
            self.temporal_storage.clear()



# -*- coding: utf-8 -*-
"""
    kaylee.session
    ~~~~~~~~~~~~~~

    This module provides base interface for managing session data
    and contains its basic implementations.

    :copyright: (c) 2012 by Zaur Nasibov.
    :license: MIT, see LICENSE for more details.
"""
#Pylint false alarm of missing hashlib functions
#pylint: disable-msg=E0611

import random
import cPickle as pickle
from base64 import b64encode, b64decode
from hmac import new as hmac
from hashlib import sha1, sha256
from Crypto.Cipher import AES
from abc import ABCMeta, abstractmethod

from .util import get_secret_key
from .errors import KayleeError

SESSION_DATA_ATTRIBUTE = '__kl_sd__'

class SessionDataManager(object):
    """The abstract base class representing Session data manager
    interface.

    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def store(self, node, task):
        """Stores the session variables found in task and then  removes
        these variables from the task, e.g. the ``#s1`` and ``#s2``
        session variables should be removed in the following task::

          task = {
              'id': 'i1',
              '#s1': 10,
              '#s2': [1, 2, 3]
          }

          manager.store(node, task)

          # task -> {'id': 'i1'}.


        :param node: the node to which the current task will be sent.
        :param task: task with or without session data. The task is accessed
                     by reference and is modified in-place.
        :type node: :class:`NodeID`
        :type task: dict
        """

    @abstractmethod
    def restore(self, node, result):
        """Restores the session variables and attaches them to the result.

        :param node: the node from which the result was received.
        :param result: the result to which the session data is attached. The
                       result is accessed by reference and is modified
                       in-place.
        :type node: :class:`NodeID`
        :type result: dict
        """

    @staticmethod
    def get_session_data(task):
        """Returns a dict with session variables found in task."""
        return { key : task[key] for key in task
                 if key.startswith('#') }

    @staticmethod
    def remove_session_data_from_task(session_data_keys, task):
        for key in session_data_keys:
            del task[key]


class PhonySessionDataManager(SessionDataManager):
    """The default session data manager which throws :class:`KayleeError`
    if any session variables are encountered in an outgoing task."""
    def store(self, node, task):
        session_data = self.get_session_data(task)
        if session_data == {}:
            return
        else:
            raise KayleeError('Cannot store session data in a '
                              'phony data manager')

    def restore(self, node, result):
        pass

KL_LOADER_DEFAULT_SESSION_DATA_MANAGER = PhonySessionDataManager


class NodeSessionDataManager(SessionDataManager):
    """A session data manager, which keeps the data in
    :attr:`Node.session_data`."""
    def store(self, node, task):
        session_data = self.get_session_data(task)
        if session_data == {}:
            return

        node.session_data = pickle.dumps(session_data, pickle.HIGHEST_PROTOCOL)
        self.remove_session_data_from_task(session_data.iterkeys(), task)

    def restore(self, node, result):
        if node.session_data is None:
            return
        session_data = pickle.loads(node.session_data)
        node.session_data = None
        result.update(session_data)


class JSONSessionDataManager(SessionDataManager):
    """Stores encrypted session variables in task and restores them
    from the results. For example, the following task data::

      task = {
          'id' : 'i1',
          '#s1' : 10,
          '#s2' : [1, 2, 3]
      }

    Encrypted via the "abc" secret key turns into::

      task = {
          id: 'i1',
          '__kl_sd__': 'yn/fCyEcW8AFrPps7XoxunC...' # 143 chars in total
      }

    The Kaylee client-side engine automatically attaches the ``'__kl_sd__``
    data to the JSON result sent to the server, so that the session data
    could be decrypted and restored, e.g.::

      {
          'speed': 20,
          '#s1': 10,
          '#s2': [1, 2, 3]
      }

    :param secret_key: An override of the global :config:`SECRET_KEY`
                       parameter.
    """
    def __init__(self, secret_key=None):
        #pylint: disable-msg=W0231
        #W0231: __init__ method from base class 'SessionDataManager'
        #       is not called.
        self._secret_key = secret_key
        self.SESSION_DATA_ATTRIBUTE = SESSION_DATA_ATTRIBUTE

    def store(self, node, task):
        session_data = self.get_session_data(task)
        if session_data == {}:
            return

        task[self.SESSION_DATA_ATTRIBUTE] =  _encrypt(session_data,
                                                      self.secret_key)
        self.remove_session_data_from_task(session_data.iterkeys(), task)

    def restore(self, node, result):
        if self.SESSION_DATA_ATTRIBUTE not in result:
            return
        sd = _decrypt(result[self.SESSION_DATA_ATTRIBUTE], self.secret_key)
        del result[self.SESSION_DATA_ATTRIBUTE]
        result.update(sd)

    @property
    def secret_key(self):
        if self._secret_key is None:
            self._secret_key = get_secret_key()
        return self._secret_key


def _encrypt(data, secret_key):
    """Encrypt the data and return its base64 representation.

    :param data: Data to encrypt. The data is pickled prior to encryption.
    :param secret_key: A secret key to use.
    :type data: any pickable Python object
    :type secret_key: str
    """
    mac = hmac(secret_key, None, sha1)
    encryption_key = sha256(secret_key).digest()

    iv = ''.join(chr(random.randint(0, 0xFF)) for i in xrange(16))
    encryptor = AES.new(encryption_key, AES.MODE_CBC, iv)

    b64_iv = b64encode(iv)
    result = [b64_iv]      # store initialization vector
    for key, val in data.iteritems():
        result.append(_encrypt_attr(key, val, encryptor))
        mac.update('|' + result[-1])

    return '{}?{}'.format(b64encode(mac.digest()), '&'.join(result))


def _decrypt(s, secret_key):
    base64_hash, data = s.split('?', 1)
    mac = hmac(secret_key, None, sha1)

    iv, data = data.split('&', 1)
    iv = b64decode(iv)

    decryption_key = sha256(secret_key).digest()
    decryptor = AES.new(decryption_key, AES.MODE_CBC, iv)

    res = {}
    for item in data.split('&'):
        mac.update('|' + item)
        attr, val = _decrypt_attr(item, decryptor)
        res[attr] = val

    if b64decode(base64_hash) == mac.digest():
        return res
    else:
        raise KayleeError('Encrypted data signature verification failed.')


def _encrypt_attr(attr, value, encryptor):
    BLOCK_SIZE = 32
    PADDING = ' '
    # one-liner to sufficiently pad the text to be encrypted
    pad = lambda s: s + (BLOCK_SIZE - len(s) % BLOCK_SIZE) * PADDING

    val = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
    val = '{}={}'.format(attr, val)
    val = encryptor.encrypt(pad(val))
    val = b64encode(val)
    return val


def _decrypt_attr(data, decryptor):
    tdata = b64decode(data)
    tdata = decryptor.decrypt(tdata).rstrip(' ')
    attr, val = tdata.split('=', 1)
    val = pickle.loads(val)
    return attr, val



from kaylee import Project
from kaylee.project import AUTO_PROJECT_MODE
from kaylee.errors import InvalidResultError

class AutoTestProject(Project):
    RESULT_KEY = 'res'
    TASKS_COUNT = 10

    def __init__(self, *args, **kwargs):
        super(AutoTestProject, self).__init__(script_url='',
                                              mode=AUTO_PROJECT_MODE,
                                              *args,
                                              **kwargs)
        self.task_id = 0
        self.tasks_count = kwargs.get('tasks_count', self.TASKS_COUNT)
        self.client_config.update({'test_key' : 'test_value'})

    def __getitem__(self, task_id):
        return {
            'id': str(task_id),
        }

    def next_task(self):
        if self.task_id < self.tasks_count:
            self.task_id += 1
            return self[self.task_id]
        else:
            return None

    def normalize_result(self, task_id, result):
        try:
            res = int(result[self.RESULT_KEY])
            return res
        except (KeyError, ValueError):
            raise InvalidResultError(result, 'The result is wrong.')

    #pylint: disable-msg=W0613
    #W0613: Unused argument 'result'
    def result_stored(self, task_id, result, storage):
        if len(storage) == self.TASKS_COUNT:
            self.completed = True



# -*- coding: utf-8 -*-
import json

from kaylee.testsuite import KayleeTest, load_tests
from kaylee import NodeID, loader

from datetime import datetime

class KayleeTests(KayleeTest):
    def setUp(self):
        self.settings = __import__('test_settings')

    def test_register_unregister(self):
        kl = loader.load(self.settings)
        node_json_config = kl.register('127.0.0.1')
        node_config = json.loads(node_json_config)
        self.assertEqual(len(node_config), 3)
        self.assertIn('node_id', node_config)
        self.assertIn('config', node_config)
        self.assertIn('applications', node_config)

        nid = NodeID(node_id = node_config['node_id'])
        self.assertIn(nid, kl.registry)
        self.assertIn(node_config['node_id'], kl.registry)

        kl.unregister(nid)
        self.assertNotIn(nid, kl.registry)

    def test_subscribe_unsubscribe(self):
        kl = loader.load(self.settings)
        app = kl.applications['test.1']
        node_json_config = kl.register('127.0.0.1')
        node_config = json.loads(node_json_config)
        node_id = node_config['node_id']

        # test node.subscribe
        app_json_config = kl.subscribe(node_id, 'test.1')
        app_config = json.loads(app_json_config)
        self.assertEqual(app_config['test_key'], 'test_value')
        node = kl.registry[node_id]
        self.assertEqual(node.controller, app)
        self.assertTrue(0 <= (datetime.now() -
                              node.subscription_timestamp).seconds < 1)

        # test node.unsubscribe
        kl.unsubscribe(node_id)
        self.assertIsNone(node.controller)
        self.assertIsNone(node.subscription_timestamp)
        self.assertIn(node, kl.registry)


kaylee_suite = load_tests([KayleeTests])
